import json as js
import logging
from datetime import datetime
from powermon.formats.abstractformat import AbstractFormat
from powermon.commands.result import Result
from powermon.commands.reading import Reading

log = logging.getLogger("hass")


class hass(AbstractFormat):
    def __init__(self, formatConfig, device):
        super().__init__(formatConfig)
        self.name = "hass"
        self.extra_info = formatConfig.get("extra_info", False)
        self.discovery_prefix = formatConfig.get("discovery_prefix", "homeassistant")
        self.entity_id_prefix = formatConfig.get("entity_id_prefix", "mpp")
        if device is None:
            self.device_name = "MPP Solar"
            self.device_id = "mpp-solar"
            self.device_model = "MPP Solar"
            self.device_manufacturer = "MPP Solar"
        else:
            self.device_name = device.name
            self.device_id = device.device_id
            self.device_model = device.model
            self.device_manufacturer = device.manufacturer

    def sendsMultipleMessages(self) -> bool:
        return True
    
    def set_command_description(self, command_description):
        pass

    def format(self, result: Result) -> list:
        log.info("Using output formatter: %s" % self.name)

        config_msgs = []
        value_msgs = []

        _result = []
        if result.readings is None:
            return _result
        display_data : list[Reading] = self.format_and_filter_data(result)
        log.debug(f"displayData: {display_data}")

        # build data to display
        for response in display_data:
            # Get key data
            data_name = response.get_data_name()
            value = response.get_data_value()
            unit = response.get_data_unit()
            icon = response.get_icon()
            device_class = response.get_device_class()
            state_class = response.get_state_class()

            # Set component type
            if unit == "bool" or value == "enabled" or value == "disabled":
                component = "binary_sensor"
            else:
                component = "sensor"

            # Make value adjustments
            if component == "binary_sensor":
                if value == 0 or value == "0" or value == "disabled":
                    value = "OFF"
                elif value == 1 or value == "1" or value == "enabled":
                    value = "ON"


            # Object ID
            object_id = f"{self.entity_id_prefix}_{data_name}".lower()

            name = f"{self.entity_id_prefix} {data_name}"

            # Home Assistant MQTT Auto Discovery Message
            #
            # Topic
            # <discovery_prefix>/<component>/[<node_id>/]<object_id>/config, eg homeassistant/binary_sensor/garden/config
            topic_base = f"{self.discovery_prefix}/{component}/{object_id}".replace(" ", "_")
            topic = f"{topic_base}/config"
            state_topic = f"{topic_base}/state"

            # Payload
            # msg '{"name": "garden", "device_class": "motion", "state_topic": "homeassistant/binary_sensor/garden/state", "unit_of_measurement": "°C", "icon": "power-plug"}'
            payload = {
                "name": f"{name}",
                "state_topic": f"{state_topic}",
                "unique_id": f"{object_id}",
                "force_update": "true",
                "last_reset": str(datetime.now()),
            }

            # Add device info
            # payload["device"] = {"name": f"{device_name}", "identifiers": ["mppsolar"], "model": "PIP6048MAX", "manufacturer": "MPP-Solar"}
            payload["device"] = {
                "name": self.device_name,
                "identifiers": [self.device_id],
                "model": self.device_model,
                "manufacturer": self.device_manufacturer,
            }

            # Add unit of measurement
            if unit and unit != "bool":
                payload["unit_of_measurement"] = f"{unit}"

            # Add icon
            if icon:
                payload.update({"icon": icon})

            # Add device class
            if device_class:
                payload["device_class"] = device_class

            # Add state_class
            if state_class:
                payload["state_class"] = state_class

            payloads = js.dumps(payload)
            # print(payloads)
            msg = {"topic": topic, "payload": payloads}
            config_msgs.append(msg)

            # VALUE SETTING
            msg = {"topic": state_topic, "payload": value}
            value_msgs.append(msg)

        # order value msgs after config to allow HA time to build entity before state data arrives
        return config_msgs + value_msgs



import logging

import requests

from .prom import prom
from ..helpers import get_kwargs

log = logging.getLogger("prom")


class prom_push(prom):
    push_url = ""

    def __str__(self):
        return "pushes Node exporter Prometheus format to PushGateway"

    def output(self, *args, **kwargs):
        # We override the method only to get the PushGateway URL from the config
        self.push_url = kwargs["push_url"]

        return super().output(*args, **kwargs)

    def handle_output(self, content: str) -> None:
        with requests.post(self.push_url, data=content) as req:
            log.debug(f"POST'ed data to PushGateway {self.push_url!r}: {req=}")
    


import logging
import re

from .baseoutput import baseoutput
from ..helpers import get_kwargs, key_wanted, pad, get_max_response_length

log = logging.getLogger("boxdraw")


class boxdraw(baseoutput):
    def __str__(self):
        return "outputs the results to standard out in a table formatted with line art boxes"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"processor.boxdraw __init__ kwargs {kwargs}")

    def printHeader(command, description):
        pass

    def output(self, *args, **kwargs):
        log.info("Using output processor: boxdraw")
        log.debug(f"kwargs {kwargs}")
        data = get_kwargs(kwargs, "data")
        if data is None:
            return

        # check if config supplied
        config = get_kwargs(kwargs, "config")
        if config is not None:
            log.debug(f"config: {config}")
            # get formatting info
            remove_spaces = config.get("remove_spaces", True)
            keep_case = config.get("keep_case", False)
            filter = config.get("filter", None)
            excl_filter = config.get("excl_filter", None)
        else:
            # get formatting info
            remove_spaces = True
            keep_case = get_kwargs(kwargs, "keep_case")
            filter = get_kwargs(kwargs, "filter")
            excl_filter = get_kwargs(kwargs, "excl_filter")

        if filter is not None:
            filter = re.compile(filter)
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)

        # remove raw response
        data.pop("raw_response", None)

        # build header
        command = data.pop("_command", "Unknown command")
        description = data.pop("_command_description", "No description found")

        # build data to display
        displayData = {}
        for key, _values in data.items():
            # remove spaces
            if remove_spaces:
                key = key.replace(" ", "_")
            if not keep_case:
                # make lowercase
                key = key.lower()
            if key_wanted(key, filter, excl_filter):
                displayData[key] = _values
        log.debug(f"displayData: {displayData}")

        # Determine column widths
        _pad = 1
        # Width of parameter column
        width_p = get_max_response_length(displayData) + _pad
        if width_p < 9 + _pad:
            width_p = 9 + _pad
        # Width of value column
        width_v = get_max_response_length(data.values()) + _pad
        if width_v < 6 + _pad:
            width_v = 6 + _pad
        # Width of units column
        width_u = get_max_response_length(data.values(), 1) + _pad
        if width_u < 5 + _pad:
            width_u = 5 + _pad
        # Total line length
        line_length = width_p + width_v + width_u + 7
        # Check if command / description line is longer and extend line if needed
        cmd_str = f" Command: {command} - {description}"
        if line_length < (len(cmd_str) + 7):
            line_length = len(cmd_str) + 7
        # Check if columns too short and expand units if needed
        if (width_p + width_v + width_u + 7) < line_length:
            width_u = line_length - (width_p + width_u + 7) - 1

        # print header
        print("\u2554" + ("\u2550" * (line_length - 2)) + "\u2557")
        print(f"\u2551{cmd_str}" + (" " * (line_length - len(cmd_str) - 2)) + "\u2551")

        # print separator
        print("\u2560" + ("\u2550" * (width_p + 1)) + "\u2564" + ("\u2550" * (width_v + 1)) + "\u2564" + ("\u2550" * (width_u + 1)) + "\u2563")
        # print column headings
        print(f"\u2551 {pad('Parameter', width_p)}\u2502 {pad('Value', width_v)}\u2502 {pad('Unit', width_u)}\u2551")
        # print separator
        print("\u255f" + ("\u2500" * (width_p + 1)) + "\u253c" + ("\u2500" * (width_v + 1)) + "\u253c" + ("\u2500" * (width_u + 1)) + "\u2562")

        # print data
        for key, values in displayData.items():
            value = values[0]
            unit = values[1]
            print(f"\u2551 {pad(key, width_p)}\u2502 {pad(value, width_v)}\u2502 {pad(unit, width_u)}\u2551")

        # print footer
        print("\u255a" + ("\u2550" * (width_p + 1)) + "\u2567" + ("\u2550" * (width_v + 1)) + "\u2567" + ("\u2550" * (width_u + 1)) + "\u255d")
        print("\n")



""" formats / abstractformat.py """
from abc import ABC, abstractmethod
import logging
import re

from powermon.dto.formatDTO import FormatDTO
from powermon.commands.result import Result
from powermon.commands.reading import Reading


# from time import sleep
log = logging.getLogger("Formatter")


def pad(text, length):
    """ expand supplied value to 'length' with spaces, return unchanged if already longer """
    if isinstance(text, float) or isinstance(text, int):
        text = str(text)
    if len(text) > length:
        return text
    return text.ljust(length, " ")


def get_max_response_lengths(readings: list[Reading]):
    """ find the max length from a list of Readings """
    _max_name = 0
    _max_value = 0
    _max_unit = 0
    for reading in readings:
        _max_name = max(len(reading.get_data_name()), _max_name)
        _max_value = max(len(str(reading.get_data_value())), _max_value)
        _max_unit = max(len(reading.get_data_unit()), _max_unit)
    return _max_name, _max_value, _max_unit


class AbstractFormat(ABC):
    def __str__(self):
        return f"Format: {self.name}"

    def __init__(self, formatConfig):
        self.name = "AbstractFormat"
        self.remove_spaces = formatConfig.get("remove_spaces", True)
        self.keep_case = formatConfig.get("keep_case", False)

        self._keyFilter = None
        _keyFilterString = formatConfig.get("filter", None)
        if _keyFilterString is not None:
            self._keyFilter = re.compile(_keyFilterString)

        self._keyExclusionfilter = None
        _keyExclusionFilterString = formatConfig.get("excl_filter", None)
        if _keyExclusionFilterString is not None:
            self._keyExclusionfilter = re.compile(_keyExclusionFilterString)

    @abstractmethod
    def set_command_description(self, command_description):
        pass

    @abstractmethod
    def format(self, result: Result) -> list:
        pass

    def to_dto(self) -> FormatDTO:
        return FormatDTO(type=self.name)
    
    # Override this if the format sends multiple messages
    def sendsMultipleMessages(self) -> bool:
        return False

    def format_and_filter_data(self, result: Result) -> list[Reading]:

        display_data = []
        for response in result.get_responses():
            if response is None:
                raise ValueError("response cannot be None")
            formatted_key = self.formatKey(response.get_data_name())
            if self.isKeyWanted(formatted_key):
                display_data.append(response)
        return display_data

    def formatKey(self, key) -> str:
        if self.remove_spaces:
            key = key.replace(" ", "_")
        if not self.keep_case:
            key = key.lower()
        return key

    def isKeyWanted(self, key) -> bool:
        # remove any specifically excluded keys
        if self._keyExclusionfilter is not None and self._keyExclusionfilter.search(key):
            # log.debug(f"key_wanted: key {key} matches excl_filter {excl_filter} so key excluded")
            return False
        if self._keyFilter is None:
            # log.debug(
            #    f"key_wanted: No filter and key {key} not excluded by excl_filter {excl_filter} so key wanted"
            # )
            return True
        elif self._keyFilter.search(key):
            # log.debug(
            #    f"key_wanted: key {key} matches filter {filter} and not excl_filter {excl_filter} so key wanted"
            # )
            return True
        else:
            # log.debug(f"key_wanted: key {key} does not match filter {filter} so key excluded")
            return False



""" commands / command_definition.py """
import logging
import re

from powermon.commands.parameter import Parameter
# from powermon.commands.reading import Reading
from powermon.commands.reading_definition import ReadingDefinition
from powermon.commands.result import ResultType
from powermon.dto.command_definition_dto import CommandDefinitionDTO

log = logging.getLogger("CommandDefinition")


class CommandDefinition:
    """ object the contains the definition and other metadata about a command """
    def __str__(self):
        return f"{self.code=}, {self.description=}, {self.result_type=}"

    def __init__(self, code, description, help_text: str, result_type : ResultType,
                 reading_definitions, parameters, test_responses: list, regex: str):
        if reading_definitions is None or len(reading_definitions) == 0:
            raise ValueError(f"reading definitions cannot be None for command_code: {code}")
        # if test_responses is None or len(test_responses) == 0:
        #     raise ValueError(f"test_responses cannot be None for command_code: {code}")
        self.code = code
        self.description = description
        self.help_text = help_text
        self.result_type : ResultType = result_type
        self.reading_definitions : dict[int, ReadingDefinition] = reading_definitions
        self.parameters : dict[str, Parameter] = parameters
        self.test_responses : list[bytes] = test_responses
        self.regex : str | None = regex

    def to_dto(self) -> CommandDefinitionDTO:
        """ convert command definition object to data transfer object """
        return CommandDefinitionDTO(
            command_code=self.code,
            description=self.description,
            help_text=self.help_text,
            result_type=str(self.result_type),
            # responses=self.response_definitions, #TODO: make DTOs for the response definitions
            # test_responses=self.test_responses,
            regex=self.regex
        )

    def is_command_code_valid(self, command_code : str) -> bool:
        """ determines if a command code is valid """
        if self.regex is None:
            return self.code == command_code
        return re.match(self.regex, command_code) is not None

    def get_type(self) -> ResultType:
        """ return the command definition type """
        return self.result_type

    def get_response_definition_count(self) -> int:
        """ return the number of reading definitions """
        if self.reading_definitions is None:
            return 0
        return len(self.reading_definitions)

    def set_parameter_value(self, parameter_value: str):
        """ sets the parameter value """
        if self.parameters is None:
            raise ValueError(f"Parameters not defined but parameter value set for command_code: {self.code}, \
                Check the command definitions")

        for parameter in self.parameters.values():
            parameter.set_value(parameter_value)
            return

    @classmethod
    def from_config(cls, protocol_dictionary : dict) -> "CommandDefinition":
        """ build command definition object from config dict """
        code = protocol_dictionary.get("name")
        description = protocol_dictionary.get("description")
        help_text = protocol_dictionary.get("help_text")
        result_type = protocol_dictionary.get("result_type")  # QUESTION: this where ResultType.ACK logic could differ
        reading_definitions : dict[int, ReadingDefinition] = \
            ReadingDefinition.multiple_from_config(protocol_dictionary.get("reading_definitions"))
        parameters : dict[str, Parameter] = Parameter.multiple_from_config(protocol_dictionary.get("parameters"))
        test_responses = protocol_dictionary.get("test_responses")
        regex = protocol_dictionary.get("regex", None)
        log.debug("code: %s description: %s reading_definitions: %s", code, description, reading_definitions)
        return cls(
            code=code, description=description, help_text=help_text, result_type=result_type,
            reading_definitions=reading_definitions, parameters=parameters, test_responses=test_responses,
            regex=regex
        )



import logging
from typing import Tuple

from .abstractprotocol import AbstractProtocol
from .protocol_helpers import vedHexChecksum

# from .pi30 import COMMANDS

log = logging.getLogger("ved")

# (AAA BBB CCC DDD EEE
# (000 001 002 003 004

COMMANDS = {
    "vedtext": {
        "name": "vedtext",
        "description": "VE Direct Text",
        "help": " -- the output of the VE Direct text protocol",
        "type": "VEDTEXT",
        "response_type": "KEYED",
        "response": [
            ["V", "Main or channel 1 battery voltage", "V", "float:r/1000"],
            ["V2", "Channel 2 battery voltage", "V", "float:r/1000"],
            ["V3", "Channel 3 battery voltage", "V", "float:r/1000"],
            ["VS", "Auxiliary starter voltage", "V", "float:r/1000"],
            ["VM", "Mid-point voltage of the battery bank", "V", "float:r/1000"],
            ["DM", "Mid-point deviation of the battery bank", "‰", "float"],
            ["VPV", "Panel voltage", "V", "float:r/1000"],
            ["PPV", "Panel power", "W", "float"],
            ["I", "Main or channel 1 battery current", "A", "float:r/1000"],
            ["I2", "Channel 2 battery current", "A", "float:r/1000"],
            ["I3", "Channel 3 battery current", "A", "float:r/1000"],
            ["IL", "Load current", "A", "float:r/1000"],
            ["LOAD", "Load output state ON/OFF", "", "bytes:r.decode()"],
            ["T", "Battery temperature", "°C", "float"],
            ["P", "Instantaneous power", "W", "float"],
            ["CE", "Consumed Amp Hours", "Ah", "float:r/1000"],
            ["SOC", "State-of-charge", "%", "float:r/10"],
            ["TTG", "Time-to-go", "Minutes", "float"],
            ["Alarm", "Alarm condition active", "", "bytes:r.decode()"],
            ["Relay", "Relay state", "", "bytes:r.decode()"],
            ["AR", "Alarm reason", "", "bytes:r.decode()"],
            ["OR", "Off reason", "", "bytes:r.decode()"],
            ["H1", "Depth of the deepest discharge", "Ah", "float:r/1000"],
            ["H2", "Depth of the last discharge", "Ah", "float:r/1000"],
            ["H3", "Depth of the average discharge", "Ah", "float:r/1000"],
            ["H4", "Number of charge cycles", "", "bytes:r.decode()"],
            ["H5", "Number of full discharges", "", "bytes:r.decode()"],
            ["H6", "Cumulative Amp Hours drawn", "Ah", "float:r/1000"],
            ["H7", "Minimum main battery voltage", "V", "float:r/1000"],
            ["H8", "Maximum main battery voltage", "V", "float:r/1000"],
            ["H9", "Number of seconds since last full charge", "Seconds", "float"],
            ["H10", "Number of automatic synchronizations", "", "bytes:r.decode()"],
            ["H11", "Number of low main voltage alarms", "", "bytes:r.decode()"],
            ["H12", "Number of high main voltage alarms", "", "bytes:r.decode()"],
            ["H13", "Number of low auxiliary voltage alarms", "", "bytes:r.decode()"],
            ["H14", "Number of high auxiliary voltage alarms", "", "bytes:r.decode()"],
            ["H15", "Minimum auxiliary battery voltage", "V", "float:r/1000"],
            ["H16", "Maximum auxiliary battery voltage", "V", "float:r/1000"],
            ["H17", "Amount of discharged energy", "kWh", "float:r/100"],
            ["H18", "Amount of charged energy", "kWh", "float:r/100"],
            ["H19", "Yield total - user resettable counter", "kWh", "float:r/100"],
            ["H20", "Yield today", "kWh", "float:r/100"],
            ["H21", "Maximum power today", "W", "float"],
            ["H22", "Yield yesterday", "kWh", "float:r/100"],
            ["H23", "Maximum power yesterday", "W", "float"],
            ["ERR", "Error code", "", "bytes:r.decode()"],
            ["CS", "State of operation", "", "bytes:r.decode()"],
            ["BMV", "Model description", "", "bytes:r.decode()"],
            ["FW", "Firmware version 16 bit", "", "bytes:r.decode()"],
            ["FWE", "Firmware version 24 bit", "", "bytes:r.decode()"],
            ["PID", "Product ID", "", "bytes:r.decode()"],
            ["SER#", "Serial number", "", "bytes:r.decode()"],
            ["HSDS", "Day sequence number 0..364", "", "bytes:r.decode()"],
            ["MODE", "Device mode", "", "bytes:r.decode()"],
            ["AC_OUT_V", "AC output voltage", "V", "float:r*100"],
            ["AC_OUT_I", "AC output current", "0.1 A", "float"],
            ["AC_OUT_S", "AC output apparent power", "VA", "float"],
            ["WARN", "Warning reason", "", "bytes:r.decode()"],
            ["MPPT", "Tracker operation mode", "", "bytes:r.decode()"],
            ["Checksum", "Checksum", "", "exclude"],
        ],
        "test_responses": [
            b"H1\t-32914\r\nH2\t0\r\nH3\t0\r\nH4\t0\r\nH5\t0\r\nH6\t-35652\r\nH7\t12041\r\nH8\t14282\r\nH9\t0\r\nH10\t0\r\nH11\t0\r\nH12\t0\r\nH15\t-22\r\nH16\t0\r\nH17\t46\r\nH18\t48\r\nChecksum\t\x1a\r\nPID\t0xA389\r\nV\t12865\r\nVS\t-14\r\nI\t0\r\nP\t0\r\nCE\t0\r\nSOC\t1000\r\nTTG\t-1\r\nAlarm\tOFF\r\nAR\t0\r\nBMV\tSmartShunt 500A/50mV\r\nFW\t0405\r\nChecksum\tL\r\n",
            b"\x00L\r\nH1\t-32914\r\nH2\t0\r\nH3\t0\r\nH4\t0\r\nH5\t0\r\nH6\t-35652\r\nH7\t12041\r\nH8\t14282\r\nH9\t0\r\nH10\t0\r\nH11\t0\r\nH12\t0\r\nH15\t-22\r\nH16\t0\r\nH17\t46\r\nH18\t48\r\nChecksum\t\x1a\r\nPID\t0xA389\r\nV\t12868\r\nVS\t-12\r\nI\t0\r\nP\t0\r\nCE\t0\r\nSOC\t1000\r\nTTG\t-1\r\nAlarm\tOFF\r\nAR\t0\r\nBMV\tSmartShunt 500A/50mV\r\nFW\t0405\r\nChecksum\tK\r",
            b"\nH1\t-32914\r\nH2\t0\r\nH3\t0\r\nH4\t0\r\nH5\t0\r\nH6\t-35652\r\nH7\t12041\r\nH8\t14282\r\nH9\t0\r\nH10\t0\r\nH11\t0\r\nH12\t0\r\nH15\t-22\r\nH16\t0\r\nH17\t46\r\nH18\t48\r\nChecksum\t\x1a\r\nPID\t0xA389\r\nV\t12868\r\nVS\t-13\r\nI\t0\r\nP\t0\r\nCE\t0\r\nSOC\t1000\r\nTTG\t-1\r\nAlarm\tOFF\r\nAR\t0\r\nBMV\tSmartShunt 500A/50mV\r\nFW\t0405\r\nChecksum\tJ\r",
        ],
    },
    "batteryCapacity": {
        "name": "batteryCapacity",
        "description": "battery capacity",
        "help": " -- display the battery capacity setting value",
        "type": "VEDGET",
        "command_code": "0010",  # or should be the more accurate 1000
        "response_type": "POSITIONAL",
        "response": [
            ["discard", 1, "Command type", ""],
            ["discard", 2, "Command", ""],
            [
                "keyed",
                1,
                "Command response flag",
                {
                    "00": "OK",
                    "01": "Unknown ID",
                    "02": "Not supported",
                    "04": "Parameter Error",
                },
            ],
            ["LittleHex2Short", 2, "Battery Capacity", "Ah"],
            ["discard", 1, "checksum", ""],
        ],
        "test_responses": [
            b":70010007800C6\n",
            b"\x00\x1a:70010007800C6\n",
        ],
    },
}


class ved(AbstractProtocol):
    """
    VED - VEDirect protocol handler
    """

    def __str__(self):
        return "VED protocol handler for Victron direct SmartShunts"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"VED"
        self.COMMANDS = COMMANDS
        self.STATUS_COMMANDS = [
            "vedtext",
        ]
        self.SETTINGS_COMMANDS = [
            "",
        ]
        self.DEFAULT_COMMAND = "vedtext"

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different for VEDirect
        """
        log.info(
            f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands"
        )
        # These need to be set to allow other functions to work`
        self._command = command
        self._command_defn = self.get_command_defn(command)
        # End of required variables setting
        if self._command_defn is None:
            return None

        # VEDHEX
        # : start of command
        # 7 Get
        # 0000 id of the value to get
        # 00 flags
        # 00 cs
        # \n
        # eg b':70010003E\n' = get battery capacity id = 0x1000 = 0010 little endian
        cmd_type = self._command_defn["type"]
        if cmd_type == "VEDTEXT":
            # Just listen - dont need to send a command
            log.debug(f"command is VEDTEXT type so returning {cmd_type}")
            return cmd_type
        elif cmd_type == "VEDGET":
            ID = self._command_defn["command_code"]
            cmd = f"7{ID}00"
            # pad cmd and convert to bytes for checksum
            _r = f"0{cmd}"
            _r = bytes.fromhex(_r)
            checksum = vedHexChecksum(_r)
            cmd = f":{cmd}{checksum:02X}\n"
            log.debug(f"full command: {cmd}")
            return cmd
        log.warning("unable to generate full command - is the definition wrong?")
        return None

    def check_response_valid(self, response) -> Tuple[bool, dict]:
        """
        VED HEX protocol - sum of bytes should be 0x55
        VED Text protocol - no validity check
        """
        if not response:
            return False, {"validity check": ["Error: Response was empty", ""]}
        if b":" in response:
            # HEX protocol response
            log.debug(f"checking validity of {response}")
            _r = response.split(b":")[1][:-1].decode()
            # print(f"trimmed response {_r}")
            _r = f"0{_r}"
            # print(f"padded response {_r}")
            _r = bytes.fromhex(_r)
            # print(f"bytes response {_r}")
            data = _r[:-1]
            checksum = _r[-1:][0]
            if vedHexChecksum(data) == checksum:
                log.debug(
                    f"VED Hex Checksum matches in response '{response}' checksum:{checksum}"
                )
                return True, {}
            else:
                # print("VED Hex Checksum does not match")
                return False, {
                    "validity check": [
                        f"Error: VED HEX checksum did not match for response {response}",
                        "",
                    ]
                }
        else:
            return True, {}

    def get_responses(self, response):
        """
        Override the default get_responses as its different for PI00
        """
        # remove \n
        response = response.replace(b"\n", b"")
        responses = []
        # for hex protocol responses - these contain a :
        if b":" in response:
            # trim anything before ':'
            _r = response.split(b":")[1].decode()
            # pad command (which is single char)
            _r = f"0{_r}"
            _r = bytes.fromhex(_r)
            if (
                self._command_defn is not None
                and self._command_defn["response_type"] == "POSITIONAL"
            ):
                # Have a POSITIONAL type response, so need to break it up...
                for defn in self._command_defn["response"]:
                    size = defn[1]
                    item = _r[:size]
                    responses.append(item)
                    _r = _r[size:]
                return responses
            else:
                return bytearray(response)
                # convert string hex to bytes
                _r = bytes.fromhex(_r)
                return bytearray(_r)
        else:
            # for text protocol responses
            _responses = response.split(b"\r")
            for resp in _responses:
                _resp = resp.split(b"\t")
                responses.append(_resp)
        # Trim leading '(' of first response
        # 3responses[0] = responses[0][1:]
        # Remove CRC and \r of last response
        # responses[-1] = responses[-1][:-3]
        return responses



""" formats / table.py """
import logging

from powermon.commands.reading import Reading
from powermon.commands.result import Result
from powermon.formats.abstractformat import (AbstractFormat,
                                             get_max_response_lengths, pad)

log = logging.getLogger("table")


class table(AbstractFormat):
    def __str__(self):
        return "outputs the results to standard out in a table (optionally formatted with line art boxes)"

    def __init__(self, formatConfig):
        super().__init__(formatConfig)
        self.name = "table"
        self.extra_info = formatConfig.get("extra_info", False)
        self.draw_lines = formatConfig.get("draw_lines", False)
        self.command_description = "unknown command"

    def set_command_description(self, command_description):
        self.command_description = command_description

    def format(self, result: Result) -> list[str]:
        log.info("Using output formatter: %s", self.name)

        _result = []

        # check for error in result
        #if result.error:
        #    data = {}
        #    data["Error"] = [f"Command: {result.command_code} incurred an error or errors during execution or processing", ""]
        #    data["Error Count"] = [len(result.error_messages), ""]
        #    for i, message in enumerate(result.error_messages):
        #        data[f"Error #{i}"] = [message, ""]

        if len(result.get_responses()) == 0:
            return _result

        filtered_responses: list[Reading] = self.format_and_filter_data(result)
        log.debug("displayData: %s", "\n".join((str(a) for a in filtered_responses)))

        # build header
        command_code = result.command_code

        # Determine column widths
        _pad = 1
        
        width_p, width_v, width_u = get_max_response_lengths(filtered_responses)
        # Width of parameter column
        width_p += _pad
        if width_p < 9 + _pad:
            width_p = 9 + _pad
        # Width of value column
        width_v += _pad
        if width_v < 6 + _pad:
            width_v = 6 + _pad
        # Width of units column
        width_u += _pad
        if width_u < 5 + _pad:
            width_u = 5 + _pad
        # Total line length
        line_length = width_p + width_v + width_u + 7
        # Check if command / description line is longer and extend line if needed
        cmd_str = f"Command: {command_code} - {self.command_description}"
        width_c = len(cmd_str)
        log.debug(f"{width_c=}, {line_length=}, {width_p=}, {width_v=}, {width_u=}")
        if line_length < (width_c + 7):
            line_length = width_c + 7
        # Check if columns too short and expand units if needed
        if (width_p + width_v + width_u + 7) <= line_length:
            width_u = line_length - (width_p + width_v + 7) 
        log.debug(f"{width_c=}, {line_length=}, {width_p=}, {width_v=}, {width_u=}")

        # print header
        if self.draw_lines:
            _result.append("\u2554" + ("\u2550" * (line_length - 2)) + "\u2557")
            _result.append(f"\u2551 {cmd_str}" + (" " * (line_length - len(cmd_str) - 3)) + "\u2551")
        else:
            _result.append("-" * (line_length))
            _result.append(f"{cmd_str}" + (" " * (line_length - len(cmd_str) - 2)))
            _result.append("-" * (line_length))

        # print separator
        if self.draw_lines:
            _result.append("\u2560" + ("\u2550" * (width_p + 1)) + "\u2564" + ("\u2550" * (width_v + 1)) + "\u2564" + ("\u2550" * (width_u + 1)) + "\u2563")
        # print column headings
        if self.draw_lines:
            _result.append(f"\u2551 {pad('Parameter', width_p)}\u2502 {pad('Value', width_v)}\u2502 {pad('Unit', width_u)}\u2551")
        else:
            _result.append(f"{pad('Parameter', width_p)} {pad('Value', width_v)} {pad('Unit', width_u)}")
        # print separator
        if self.draw_lines:
            _result.append("\u255f" + ("\u2500" * (width_p + 1)) + "\u253c" + ("\u2500" * (width_v + 1)) + "\u253c" + ("\u2500" * (width_u + 1)) + "\u2562")

        # print data
        for response in filtered_responses:
            name = response.get_data_name()
            value = response.get_data_value()
            unit = response.get_data_unit()
            if self.draw_lines:
                _result.append(f"\u2551 {pad(name, width_p)}\u2502 {pad(value, width_v)}\u2502 {pad(unit, width_u)}\u2551")
            else:
                _result.append(f"{pad(name, width_p)} {pad(value, width_v)} {pad(unit, width_u)}")

        # print footer
        if self.draw_lines:
            _result.append("\u255a" + ("\u2550" * (width_p + 1)) + "\u2567" + ("\u2550" * (width_v + 1)) + "\u2567" + ("\u2550" * (width_u + 1)) + "\u255d")
        # _result.append("\n")
        return _result


def get_max_response_length(responses: list[Reading]):
    """Helper function for table format"""
    _max_length = 0
    for response in responses:
        data_string = str(response.get_data_value())
        if len(data_string) > _max_length:
            _max_length = len(data_string)
    return _max_length

def pad(text, length):
    if type(text) == float or type(text) == int:
        text = str(text)
    if len(text) > length:
        return text
    return text.ljust(length, " ")


from typing import Optional
from pydantic import BaseModel, Field

from .protocolDTO import ProtocolDTO


class PortDTO(BaseModel):
    type: str
    path: str = Field(default="")
    baud: int = Field(default=9600)
    protocol: ProtocolDTO



import logging
from typing import Tuple

from .abstractprotocol import AbstractProtocol
from .protocol_helpers import crc8 as dalyChecksum

# from .pi30 import COMMANDS

log = logging.getLogger("daly")

# (AAA BBB CCC DDD EEE
# (000 001 002 003 004

COMMANDS = {
    "SOC": {
        "name": "SOC",
        "description": "State of Charge",
        "help": " -- display the battery state of charge",
        "type": "DALY",
        "command_code": "90",
        "response_type": "POSITIONAL",
        "response": [
            ["discard", 1, "start flag", ""],
            ["discard", 1, "module address", ""],
            ["discard", 1, "command id", ""],
            ["discard", 1, "data length", ""],
            ["BigHex2Short:r/10", 2, "Battery Bank Voltage", "V"],
            ["BigHex2Short:r/10", 2, "acquistion", "V"],
            ["BigHex2Short:(r-30000)/10", 2, "Current", "A"],
            ["BigHex2Short:r/10", 2, "SOC", "%"],
            ["discard", 1, "checksum", ""],
        ],
        "test_responses": [
            b"\xa5\x01\x90\x08\x02\x10\x00\x00uo\x03\xbc\xf3",
            b"\xa5\x01\x90\x08\x02\x14\x00\x00uE\x03x\x89",
            b"\xa5\x01\x90\x08\x01\t\x00\x00u\xcf\x03\n\x99",
        ],
    },
    "cellMinMaxVoltages": {
        "name": "cellMinMaxVoltages",
        "description": "Cell Minimum and Maximum Voltages",
        "help": " -- display the cell number and voltage of the cells with the min and max voltages",
        "type": "DALY",
        "command_code": "91",
        "response_type": "POSITIONAL",
        "response": [
            ["discard", 1, "start flag", ""],
            ["discard", 1, "module address", ""],
            ["discard", 1, "command id", ""],
            ["discard", 1, "data length", ""],
            ["BigHex2Short:r/1000", 2, "Maximum Cell Voltage", "V"],
            ["Hex2Int", 1, "Maximum Cell Number", ""],
            ["BigHex2Short:r/1000", 2, "Minimum Cell Voltage", "V"],
            ["Hex2Int", 1, "Minimum Cell Number", ""],
            ["discard", 2, "SOC", "%"],
            ["discard", 1, "checksum", ""],
        ],
        "test_responses": [
            b"\xa5\x01\x91\x08\r\x00\x0f\x0c\xfe\x01\x03x\xe1",
        ],
    },
    "cellMinMaxTemps": {
        "name": "cellMinMaxTemps",
        "description": "Cell Minimum and Maximum Temperatures",
        "help": " -- display the cell number and temperature of the cells with the min and max temperatures",
        "type": "DALY",
        "command_code": "92",
        "response_type": "POSITIONAL",
        "response": [
            ["discard", 1, "start flag", ""],
            ["discard", 1, "module address", ""],
            ["discard", 1, "command id", ""],
            ["discard", 1, "data length", ""],
            ["Hex2Int:r-40", 1, "Maximum Cell Temperature", "°C"],
            ["Hex2Int", 1, "Maximum Cell Number", ""],
            ["Hex2Int:r-40", 1, "Minimum Cell Temperature", "°C"],
            ["Hex2Int", 1, "Minimum Cell Number", ""],
            ["discard", 2, "not used", ""],
            ["discard", 2, "SOC", "%"],
            ["discard", 1, "checksum", ""],
        ],
        "test_responses": [b"\xa5\x01\x92\x087\x017\x01\xfe\x01\x03x*"],
    },
    "mosStatus": {
        "name": "mosStatus",
        "description": "mosStatus",
        "help": " -- display the mosStatus",
        "type": "DALY",
        "command_code": "93",
        "response_type": "POSITIONAL",
        "response": [
            ["discard", 1, "start flag", ""],
            ["discard", 1, "module address", ""],
            ["discard", 1, "command id", ""],
            ["discard", 1, "data length", ""],
            ["hex_option", 1, "Charge Status", ["stationary", "charged", "discharged"]],
            ["Hex2Str", 1, "Charging MOS Tube Status", ""],
            ["Hex2Str", 1, "Discharging MOS Tube Status", ""],
            ["Hex2Int", 1, "BMS Life", "cycles"],
            ["BigHex2Float:r/1000", 4, "Residual Capacity", "AH"],
            ["discard", 1, "checksum", ""],
        ],
        "test_responses": [
            b"\xa5\x01\x93\x08\x02\x01\x01x\x00\x03\xcb@\xcb",
            b"\xa5\x01\x93\x08\x00\x01\x01\x9a\x00\x02\xa2\xd8Y",
        ],
    },
    "status": {
        "name": "status",
        "description": "Status Information",
        "help": " -- display the status information",
        "type": "DALY",
        "command_code": "94",
        "response_type": "POSITIONAL",
        "response": [
            ["discard", 1, "start flag", ""],
            ["discard", 1, "module address", ""],
            ["discard", 1, "command id", ""],
            ["discard", 1, "data length", ""],
            ["Hex2Int", 1, "Number of Cells", ""],
            ["Hex2Int", 1, "Number of Temperature Sensors", ""],
            ["hex_option", 1, "Charger Status", ["disconnected", "connected"]],
            ["hex_option", 1, "Load Status", ["disconnected", "connected"]],
            ["Hex2Str", 1, "Flags (TODO)", ""],
            ["BigHex2Short", 2, "Charge/Discharge Cycles", "cycles"],
            ["Hex2Str", 1, "Reserved", ""],
            ["discard", 1, "checksum", ""],
        ],
        "test_responses": [
            b"\xa5\x01\x94\x08\x10\x01\x00\x00\x00\x00\x03@\x96",
        ],
    },
    "cellVoltages": {
        "name": "cellVoltages",
        "description": "Cell Voltages Information",
        "help": " -- display the cell voltages",
        "type": "DALY",
        "command_code": "95",
        "response_type": "MULTIFRAME-POSITIONAL",
        "response_length": 13,
        "response": [
            ["discard", 1, "start flag", ""],
            ["discard", 1, "module address", ""],
            ["discard", 1, "command id", ""],
            ["discard", 1, "data length", ""],
            ["discard", 1, "f'Frame Number {f:02d}'", ""],
            ["BigHex2Short:r/1000", 2, "f'Cell {3*f+1:02d} Voltage'", "V"],
            ["BigHex2Short:r/1000", 2, "f'Cell {3*f+2:02d} Voltage'", "V"],
            ["BigHex2Short:r/1000", 2, "f'Cell {3*f+3:02d} Voltage'", "V"],
            ["discard", 1, "Reserved", ""],
            ["discard", 1, "checksum", ""],
        ],
        "test_responses": [
            b"\xa5\x01\x95\x08\x01\x0c\xfd\x0c\xfe\x0c\xfe@\xa1\xa5\x01\x95\x08\x02\x0c\xfe\x0c\xfe\x0c\xfe@\xa3\xa5\x01\x95\x08\x03\x0c\xfe\x0c\xfe\x0c\xfe@\xa4\xa5\x01\x95\x08\x04\x0c\xfe\x0c\xfc\x0c\xfe@\xa3\xa5\x01\x95\x08\x05\x0c\xfe\x0c\xff\x0c\xfe@\xa7\xa5\x01\x95\x08\x06\x0c\xfc\x00\x00\x00\x00@\x91\xa5\x01\x95\x08\x07\x00\x00\x00\x00\x00\x00@\x8a\xa5\x01\x95\x08\x08\x00\x00\x00\x00\x00\x00@\x8b\xa5\x01\x95\x08\t\x00\x00\x00\x00\x00\x00@\x8c\xa5\x01\x95\x08\n\x00\x00\x00\x00\x00\x00@\x8d\xa5\x01\x95\x08\x0b\x00\x00\x00\x00\x00\x00@\x8e\xa5\x01\x95\x08\x0c\x00\x00\x00\x00\x00\x00@\x8f\xa5\x01\x95\x08\r\x00\x00\x00\x00\x00\x00@\x90\xa5\x01\x95\x08\x0e\x00\x00\x00\x00\x00\x00@\x91\xa5\x01\x95\x08\x0f\x00\x00\x00\x00\x00\x00@\x92\xa5\x01\x95\x08\x10\x00\x00\x00\x00\x00\x00@\x93",
            b"\xa5\x01\x95\x08\x01\x0c\xa3\x0c\xa7\x0c\xa6\xf9Q\xa5\x01\x95\x08\x02\x0c\xa8\x0c\xa5\x0c\xa4\xf9S\xa5\x01\x95\x08\x03\x0c\xa5\x0c\xa6\x00\x00\xf9\xa2\xa5\x01\x95\x08\x04\x00\x00\x00\x00\x00\x00\xf9@\xa5\x01\x95\x08\x05\x00\x00\x00\x00\x00\x00\xf9A\xa5\x01\x95\x08\x06\x00\x00\x00\x00\x00\x00\xf9B\xa5\x01\x95\x08\x07\x00\x00\x00\x00\x00\x00\xf9C\xa5\x01\x95\x08\x08\x00\x00\x00\x00\x00\x00\xf9D\xa5\x01\x95\x08\t\x00\x00\x00\x00\x00\x00\xf9E\xa5\x01\x95\x08\n\x00\x00\x00\x00\x00\x00\xf9F\xa5\x01\x95\x08\x0b",
        ],
    },
    "cellTemperatures": {
        "name": "cellTemperatures",
        "description": "Cell Temperature Information",
        "help": " -- display the cell temperatures",
        "type": "DALY",
        "command_code": "96",
        "response_type": "MULTIFRAME-POSITIONAL",
        "response_length": 13,
        "response": [
            ["discard", 1, "start flag", ""],
            ["discard", 1, "module address", ""],
            ["discard", 1, "command id", ""],
            ["discard", 1, "data length", ""],
            ["discard", 1, "f'Frame Number {f:02d}'", ""],
            ["Hex2Int:r-40", 1, "f'Cell {7*f+1:02d} Temperature'", "°C"],
            ["Hex2Int:r-40", 1, "f'Cell {7*f+2:02d} Temperature'", "°C"],
            ["Hex2Int:r-40", 1, "f'Cell {7*f+3:02d} Temperature'", "°C"],
            ["Hex2Int:r-40", 1, "f'Cell {7*f+4:02d} Temperature'", "°C"],
            ["Hex2Int:r-40", 1, "f'Cell {7*f+5:02d} Temperature'", "°C"],
            ["Hex2Int:r-40", 1, "f'Cell {7*f+6:02d} Temperature'", "°C"],
            ["Hex2Int:r-40", 1, "f'Cell {7*f+7:02d} Temperature'", "°C"],
            ["discard", 1, "checksum", ""],
        ],
        "test_responses": [
            b"\xa5\x01\x96\x08\x017\x00\x00\x00\x00\x00\x00|\xa5\x01\x96\x08\x02\x00\x00\x00\x00\x00\x00\x00F",
        ],
    },
}
startFlag = bytes.fromhex("A5")


class daly(AbstractProtocol):
    """
    DALY - Daly BMS protocol handler
    """

    def __str__(self):
        return "Daly BMS protocol handler"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"DALY"
        self.module_address = bytes.fromhex("80")
        self.COMMANDS = COMMANDS
        self.STATUS_COMMANDS = [
            "SOC",
        ]
        self.SETTINGS_COMMANDS = [
            "",
        ]
        self.DEFAULT_COMMAND = "SOC"

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different
        """
        log.info(
            f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands"
        )
        # These need to be set to allow other functions to work`
        self._command = command
        self._command_defn = self.get_command_defn(command)
        # End of required variables setting
        if self._command_defn is None:
            return None

        # DALY
        # startFlag = bytes.fromhex("A5")
        commandID = bytes.fromhex(self._command_defn["command_code"])
        dataLength = bytes.fromhex("08")
        data = bytes.fromhex("00" * 8)
        cmd = startFlag + self.module_address + commandID + dataLength + data

        checksum = f"{dalyChecksum(cmd):02X}"
        cmd = cmd + bytes.fromhex(checksum) + b"\n"
        log.debug(f"full command: {cmd}")
        return cmd

    def is_multiframe(self, response) -> bool:
        # startFlag = bytes.fromhex("A5")
        return (
            (response_length := self._command_defn.get("response_length"))
            and len(response) > response_length
        )

    def check_response_valid(self, response) -> Tuple[bool, dict]:
        """
        DALY protocol - checksum is sum of bytes
        """
        log.debug(f"checking validity of {response}")
        if not response:
            return False, {"validity check": ["Error: Response was empty", ""]}

        # Check to see if the response is a multi frame response
        if self.is_multiframe(response):
            log.info("is multiframe response - assuming ok for now")
            # TODO: fix check_response_valid for multiframe results
            return True, {}

        _r = response
        # print(f"bytes response {_r}")
        data = _r[:-1]
        checksum = _r[-1:][0]
        if dalyChecksum(data) == checksum:
            log.debug(
                f"DALY Checksum matches response '{response}' checksum:{checksum}"
            )
            return True, {}
        else:
            # print("VED Hex Checksum does not match")
            return False, {
                "validity check": [
                    f"Error: DALY checksum did not match for response {response}",
                    "",
                ]
            }

    def get_responses(self, response):
        """
        Override the default get_responses as its different
        """
        responses = []
        # remove \n
        # response = response.replace(b"\n", b"")

        if (
            self._command_defn is not None
            and self._command_defn["response_type"] == "MULTIFRAME-POSITIONAL"
        ):
            # Have multiple frames of positional data
            # Split into frames
            frame_size = self._command_defn["response_length"]
            frames = [
                response[i : i + frame_size]
                for i in range(0, len(response), frame_size)
            ]
            log.info(f"Multi frame response with {len(frames)} frames")
            # Loop through each frame and process as per definition
            for frame in frames:
                if len(frame) == 0:
                    continue
                items = []
                for defn in self._command_defn["response"]:
                    size = defn[1]
                    item = frame[:size]
                    items.append(item)
                    frame = frame[size:]
                responses.append(items)
            # print(responses)
            return responses

        if (
            self._command_defn is not None
            and self._command_defn["response_type"] == "POSITIONAL"
        ):
            # Have a POSITIONAL type response, so need to break it up...
            # example defn :
            # "response": [
            #   ["discard", 1, "start flag", ""],
            #   ["discard", 1, "module address", ""],
            #   ["discard", 1, "command id", ""],
            #   ["discard", 1, "data length", ""],
            # ]
            # example response data b"\xa5\x01\x90\x08\x02\x10\x00\x00uo\x03\xbc\xf3",
            for defn in self._command_defn["response"]:
                size = defn[1]
                item = response[:size]
                responses.append(item)
                response = response[size:]
            return responses
        else:
            return bytearray(response)



from abc import ABC, abstractmethod
import logging

# from time import sleep
log = logging.getLogger("BaseIO")


class BaseIO(ABC):
    @abstractmethod
    def send_and_receive(self, *args, **kwargs) -> dict:
        raise NotImplementedError

    def connect(self) -> None:
        log.debug("connect not implemented")
        return

    def disconnect(self) -> None:
        log.debug("disconnect not implemented")
        return

    def process_command(self, command, protocol):
        # Band-aid solution, need to reduce what is sent
        log.debug(f"Command {command}")
        full_command = protocol.get_full_command(command)

        raw_response = self.send_and_receive(
            command=command,
            full_command=full_command,
            protocol=protocol,
            command_defn=protocol.get_command_defn(command),
        )
        log.debug(f"Send and Receive Response {raw_response}")

        # Handle errors
        # Maybe there should a decode for ERRORs and WARNINGS...
        # Some inverters return the command if the command is unknown:
        if raw_response == full_command:
            return {
                "ERROR": [
                    f"Inverter returned the command string for {command} - the inverter didnt recognise this command",
                    "",
                ]
            }
        # dict is returned on exception

        if isinstance(raw_response, dict):
            return raw_response

        # Decode response
        decoded_response = protocol.decode(raw_response, command)
        log.info(f"Decoded response {decoded_response}")

        return decoded_response



""" testport.py """
import logging
import random

from powermon.dto.portDTO import PortDTO
from powermon.commands.result import Result
from powermon.ports.abstractport import AbstractPort
from powermon.protocols import get_protocol_definition
from powermon.commands.command import Command
from powermon.commands.command_definition import CommandDefinition

log = logging.getLogger("test")


class TestPort(AbstractPort):
    """ test port object - responds with test data (if configured in the protocol) """

    @classmethod
    def from_config(cls, config=None):
        log.debug("building test port. config:%s", config)
        # allows specification of which of the test responses to use (mainly to allow test cases to be repeatable)
        response_number = config.get("response_number", None)
        # get protocol handler, default to PI30 if not supplied
        protocol = get_protocol_definition(protocol=config.get("protocol", "PI30"))
        return cls(response_number=response_number, protocol=protocol)

    def __init__(self, response_number, protocol):
        super().__init__(protocol=protocol)
        self.response_number = response_number
        self.connected = False
        self._test_data = None

    def __str__(self):
        return "Test port"

    def to_dto(self) -> PortDTO:
        dto = PortDTO(type="test", protocol=self.get_protocol().to_dto())
        return dto

    def is_connected(self):
        log.debug("Test port is connected")
        return True

    def connect(self) -> int:
        log.debug("Test port connected")
        self.connected = True
        return 1

    def disconnect(self) -> None:
        log.debug("Test port disconnected")
        self.connected = False

    def send_and_receive(self, command: Command) -> Result:
        command_defn : CommandDefinition = command.command_definition

        if command_defn is not None:
            # Have test data defined, so use that
            number_of_test_responses = len(command_defn.test_responses)
            if self.response_number is not None and self.response_number < number_of_test_responses:
                self._test_data = command_defn.test_responses[self.response_number]
            else:
                self._test_data = command_defn.test_responses[random.randrange(number_of_test_responses)]
        else:
            # No test responses defined
            raise ValueError(f"Testing a command '{command.code}' with no test responses defined")
        # Get raw response
        response_line = self._test_data
        log.debug("Raw response: %s", response_line)

        # response = self.get_protocol().check_response_and_trim(response_line)
        result = command.build_result(raw_response=response_line, protocol=self.get_protocol())
        return result



""" result.py """
from enum import Enum, auto
import logging

from powermon.commands.reading import Reading
from powermon.commands.parameter import Parameter
from powermon.commands.reading_definition import ReadingDefinition, ReadingDefinitionMessage, ResponseType
from powermon.dto.resultDTO import ResultDTO

log = logging.getLogger("result")


class ResultType(Enum):
    """ enum of valid types of Results """
    ERROR = auto()
    ACK = auto()
    COMMAND = auto()
    MULTIVALUED = auto()
    INDEXED = auto()
    SINGLE = auto()
    POSITIONAL = auto()


class Result:
    """ object to contain all the info of a result """
    def __str__(self):
        return f"Result: {self.is_valid=}, {self.error=} - {self.error_messages=}, {self.raw_response=}, {' '.join(str(i) for i in self.readings)}"

    def __init__(self, command_code: str, result_type: str, raw_response: bytes, reading_definitions: list[ReadingDefinition] = None,
                 parameters: dict[str, Parameter] = None):
        if raw_response is None:
            raise ValueError("raw_response cannot be None")

        self.device_id = "default"
        self.command_code = command_code
        self.result_type = result_type
        self.raw_response = raw_response
        self.parameters = parameters
        
        self.reading_definitions = reading_definitions
        if(self.reading_definitions is None):
            reading_definition = ReadingDefinitionMessage(index=0, name="default", response_type=ResponseType.STRING , description="default")
            self.reading_definitions = [reading_definition]
            
        self.readings: list[Reading] = self.decode_response(raw_response=raw_response)
        self.is_valid = True
        self.error = False
        self.error_messages = []
        
        if result_type == ResultType.ERROR:
            self.is_valid = False
            self.error = True
            self.error_messages = [self.raw_response]
        
        log.debug("Result: %s", self)

    def to_dto(self) -> ResultDTO:
        """ convert result object to data transfer object """
        reading_dtos = []
        for reading in self.readings:
            reading_dtos.append(reading.to_dto())
        return ResultDTO(device_identifier=self.get_device_id(), command_code=self.command_code, data=reading_dtos)

    def get_command_code(self) -> str:
        return self.command_code

    def set_device_id(self, device_id):
        self.device_id = device_id

    def get_device_id(self) -> str:
        return self.device_id

    def get_responses(self) -> list[Reading]:
        return self.readings

    def add_readings(self, responses: list[Reading]) -> bool:
        self.readings.extend(responses)
        return True

    # If we split results into different types then each type can have its own decode_response
    def decode_response(self, raw_response) -> list[Reading]:
        """
        Take the raw response and decode into a list of Readings depending on the result type
        """

        log.info("result.raw_response passed to decode: %s", raw_response)

        all_readings : list[Reading] = []

        # Process response based on result type
        match self.result_type:
            case ResultType.SINGLE:
                readings = self.validate_and_translate_raw_value(self.raw_response, index=0)
                all_readings.extend(readings)
            case ResultType.ACK:
                readings = self.validate_and_translate_raw_value(self.raw_response, index=0)
                all_readings.extend(readings)
            case ResultType.INDEXED:
                # Response is splitable and order of each item determines decode logic
                for i, _raw_response in enumerate(self.split_responses(self.raw_response)):
                    print(f"i: {i}, _raw_response: {_raw_response}")
                    readings = self.validate_and_translate_raw_value(_raw_response, index=i)
                    all_readings.extend(readings)
            case ResultType.MULTIVALUED:
                # while response has multiple values, the all relate to a single result
                readings = self.validate_and_translate_raw_value(self.raw_response, index=0)
                all_readings.extend(readings)
            case ResultType.ERROR:
                readings = self.validate_and_translate_raw_value(self.raw_response, index=0)
                all_readings.extend(readings)
            case _:
                # unknown result type
                raise ValueError(f"Unknown result type: {self.result_type}")

        return all_readings

    def split_responses(self, response) -> list:
        """
        Default implementation of split and trim
        """
        # CRC should be removed by protocol, so just split split
        return response.split(None) # split differs by protocol

    def validate_and_translate_raw_value(self, raw_value: str, index: int) -> list[Reading]:
        if len(self.reading_definitions) <= index:
            log.debug("Index %s is out of range for command %s", index, self.command_code)
            reading_definition: ReadingDefinition = ReadingDefinitionMessage(index=index, name="default", response_type=ResponseType.STRING , description=f"Unused response {index}")
        else:
            reading_definition: ReadingDefinition = self.reading_definitions[index]
        try:
            return reading_definition.reading_from_raw_response(raw_value)
        except ValueError:
            error = Reading(
                data_name=reading_definition.get_description(), data_value=reading_definition.get_invalid_message(raw_value), data_unit=""
            )
            error.is_valid = False
            return [error]



from fastapi import FastAPI, Request, Depends, HTTPException
# from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates
from fastapi_mqtt import FastMQTT, MQTTConfig
# import json
# import asyncio

from sqlalchemy.orm import Session
import logging

from api.app.statehandler import StateHandler

from powermon.dto.resultDTO import ResultDTO
from powermon.dto.deviceDTO import DeviceDTO
from powermon.dto.commandDTO import CommandDTO

from .db import crud, models, schemas
from .db.database import SessionLocal, engine

log = logging.getLogger("main")

models.Base.metadata.create_all(bind=engine)

app = FastAPI()

mqtt_config = MQTTConfig(host="localhost", port=1883, username="", password="")

mqtt = FastMQTT(config=mqtt_config)

mqtt.init_app(app)


# Dependency
def get_db():
    db = SessionLocal()
    try:
        yield db
    finally:
        db.close()


def get_statehandler():
    handler = StateHandler()
    return handler


@mqtt.on_connect()
def connect(client, flags, rc, properties):
    topic = "powermon/"
    # mqtt.client.subscribe(topic) #subscribing mqtt topic
    log.info(f"Connected: {client}, {flags}, {rc}, {properties}, {topic}")



@mqtt.subscribe("powermon/announce")
async def listen_to_announcements(client, topic, payload, qos, properties):
    handler = StateHandler()
    device = handler.recieved_announcement(payload.decode())
    print(f"Recieved announcement for {device.device_id}")
    mqtt.client.subscribe(f"powermon/{device.device_id}/results/#")

@mqtt.on_message()
async def message(client, topic, payload, qos, properties):
    #print("Received message: ", topic, payload.decode(), qos, properties)
    handler = StateHandler()
    log.debug(F"Command result on topic {topic}")
    #if handler.is_command_result_topic(topic):
    handler.recieved_result(topic, payload)

@mqtt.on_disconnect()
def disconnect(client, packet, exc=None):
    print("Disconnected")


@mqtt.on_subscribe()
def subscribe(client, mid, qos, properties):
    print("subscribed", client, mid, qos, properties)


app.mount("/static", StaticFiles(directory="api/app/static"), name="static")

templates = Jinja2Templates(directory="api/app/templates")


@app.get("/")
def read_root(request: Request):
    handler = StateHandler()
    devices = handler.get_device_instances()
    print(devices)
    return templates.TemplateResponse("home.html.j2", {"request": request, "devices": devices})


@app.get("/messages/", response_model=list[schemas.MQTTMessage])
def read_messages(skip: int = 0, limit: int = 100, db: Session = Depends(get_db)):
    users = crud.get_mqtt_messages(db, skip=skip, limit=limit)
    return users


@app.get("/messages/{message_id}", response_model=schemas.MQTTMessage)
def read_message(message_id: int, db: Session = Depends(get_db)):
    mqtt_message = crud.get_mqtt_message(db, mqtt_message_id=message_id)
    if mqtt_message is None:
        raise HTTPException(status_code=404, detail="User not found")
    return mqtt_message


@app.get("/devices/{device_id}/runCommandCode/{command_code}", response_model=ResultDTO)
async def read_command(device_id: str, command_code: str, handler: StateHandler = Depends(get_statehandler)):  # noqa: F811
    result = None
    
    command_dto = CommandDTO.run_api_command(command_code=command_code, device_id=device_id)
    log.debug(f"Subcribing to {command_dto.result_topic}")
    mqtt.client.subscribe(command_dto.result_topic)
    await add_command(command_dto, handler=handler)

    result = await handler.get_next_result(command_dto.result_topic)

    return result

@app.post("/addCommand/", response_model=CommandDTO)
async def add_command(command_dto: CommandDTO, handler: StateHandler = Depends(get_statehandler)):
    log.debug("Add command")
    device = handler.get_device_instance(command_dto.device_id)
    if device is None:
        raise HTTPException(status_code=404, detail=f"Device id: {command_dto.device_id} not found")

    if command_dto.command_code not in device.port.protocol.commands.keys():
        raise HTTPException(
            status_code=404, 
            detail=f"Command: {command_dto.command_code} not found in protocol: {device.port.protocol.protocol_id}")

    mqtt.client.publish(f"powermon/{command_dto.device_id}/addcommand", command_dto.json())

    return command_dto


@app.get("/devices/", response_model=list[DeviceDTO])
def read_devices(handler: StateHandler = Depends(get_statehandler)):
    result = None

    result = handler.get_device_instances()

    return result


@app.get("/devices/{device_id}", response_model=DeviceDTO)
def read_device(device_id: str, handler: StateHandler = Depends(get_statehandler)):  # noqa: F811
    result = None

    result = handler.get_device_instance(device_id)

    if result is None:
        raise HTTPException(status_code=404, detail="Device not found")

    return result



@app.get("/results/", response_model=list[ResultDTO])
async def read_results(handler: StateHandler = Depends(get_statehandler)):
    result = None

    result = await handler.get_results()
    if result is None:
        raise HTTPException(status_code=404, detail="No command results found")

    return result



import json as js
import logging
import re
import socket

from . import to_json
from .baseoutput import baseoutput
from ..helpers import get_kwargs

log = logging.getLogger("json_udp")


class json_udp(baseoutput):
    def __str__(self):
        return "outputs all the results to tcp UDP datagram packet in JSON format"

    def __init__(self, *args, **kwargs) -> None:
        log.debug(f"__init__: kwargs {kwargs}")

    def output(self, *args, **kwargs):
        data = get_kwargs(kwargs, "data")
        tag = get_kwargs(kwargs, "tag")
        keep_case = get_kwargs(kwargs, "keep_case")
        udp_port = get_kwargs(kwargs, "udp_port", "5555")
        filter = get_kwargs(kwargs, "filter")
        if filter is not None:
            filter = re.compile(filter)
        excl_filter = get_kwargs(kwargs, "excl_filter")
        if excl_filter is not None:
            excl_filter = re.compile(excl_filter)

        msgs = []
        # Remove command and _command_description
        cmd = data.pop("_command", None)
        data.pop("_command_description", None)
        data.pop("raw_response", None)
        if tag is None:
            tag = cmd
        output = to_json(data, keep_case, excl_filter, filter)

        payload = js.dumps(output)
        log.debug(payload)
        msgs.append(payload)
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  # UDP datagram
        for msg in msgs:
            count = sock.sendto(bytes(msg, "utf-8"), ("localhost", int(udp_port)))
        log.debug(f"Udp sent response {count}")
        return msgs



# shamelessly stolen from ccrisan https://github.com/qtoggle/qtoggleserver-mppsolar/blob/master/qtoggleserver/mppsolar/io.py
import logging
import os
import time

from .baseio import BaseIO
from ..helpers import get_kwargs

log = logging.getLogger("HIDRawIO")


class HIDRawIO(BaseIO):
    def __init__(self, device_path: str) -> None:
        # self._fd = os.open(device_path, flags=os.O_RDWR | os.O_NONBLOCK)
        self._device = device_path

    def send_and_receive(self, *args, **kwargs) -> dict:
        full_command = get_kwargs(kwargs, "full_command")
        response_line = bytes()
        usb0 = None
        try:
            usb0 = os.open(self._device, os.O_RDWR | os.O_NONBLOCK)
        except Exception as e:
            log.error("USB open error: {}".format(e))
            return {"ERROR": ["USB open error: {}".format(e), ""]}
        # Send the command to the open usb connection
        cmd_len = len(full_command)
        log.debug(f"length of to_send: {cmd_len}")
        # for command of len < 8 it ok just to send
        # otherwise need to pack to a multiple of 8 bytes and send 8 at a time
        if cmd_len <= 8:
            # Send all at once
            log.debug("sending full_command in on shot")
            time.sleep(0.05)
            os.write(usb0, full_command)
        else:
            log.debug("multiple chunk send")
            chunks = [full_command[i:i + 8] for i in range(0, cmd_len, 8)]
            for chunk in chunks:
                # pad chunk to 8 bytes
                if len(chunk) < 8:
                    padding = 8 - len(chunk)
                    chunk += b'\x00' * padding
                log.debug("sending chunk: %s" % (chunk))
                time.sleep(0.05)
                os.write(usb0, chunk)
        time.sleep(0.25)
        # Read from the usb connection
        # try to a max of 100 times
        for x in range(100):
            # attempt to deal with resource busy and other failures to read
            try:
                time.sleep(0.15)
                r = os.read(usb0, 256)
                response_line += r
            except Exception as e:
                log.debug("USB read error: {}".format(e))
            # Finished is \r is in byte_response
            if bytes([13]) in response_line:
                # remove anything after the \r
                response_line = response_line[: response_line.find(bytes([13])) + 1]
                break
        log.debug("usb response was: %s", response_line)
        os.close(usb0)
        return response_line



import logging

from .pi17 import pi17


log = logging.getLogger("pi17m058")

QUERY_COMMANDS = {
    "BATS": {
        "name": "BATS",
        "description": "Query battery setting",
        "help": " -- queries battery setting",
        "type": "QUERY",
        "response_type": "INDEXED",
        "response": [
            [1, "Battery maximum charge current", "int:r/10", "A"],
            [2, "Battery constant charge voltage", "int:r/10", "V"],
            [3, "Battery floating charge voltage", "int:r/10", "V"],
            [4, "Battery stop charger current level in floating charging", "int:r/10", "A"],
            [
                5,
                "Keep charged time of battery catch stopped charging current level",
                "int",
                "Minutes",
            ],
            [
                6,
                "Battery voltage of recover to charge when battery stop charger in floating charging",
                "int:r/10",
                "V",
            ],
            [7, "Battery under voltage", "int:r/10", "V"],
            [8, "Battery under voltage release", "int:r/10", "V"],
            [9, "Battery weak voltage in hybrid mode", "int:r/10", "V"],
            [10, "Battery weak voltage release in hybrid mode", "int:r/10", "V"],
            [11, "Battery Type", "option", ["Ordinary", "Li-Fe"]],
            [12, "Reserved1", "discard", ""],
            [13, "Reserved2", "discard", ""],
            [
                14,
                "AC charger keep battery voltage function enable/diable",
                "option",
                ["Disabled", "Enabled"],
            ],
            [15, "AC charger keep battery voltage", "int:r/10", "V"],
            [16, "Battery temperature sensor compensation", "int:r/10", "mV"],
            [17, "Max. AC charging current", "int:r/10", "A"],
            [18, "Battery discharge max current in hybrid mode", "int", "A"],
            [19, "Battery under SOC", "int", "%"],
            [20, "Battery under back SOC", "int", "%"],
            [21, "Battery weak SOC in hybrid mode", "int", "%"],
            [22, "Battery weak back SOC in hybrid mode", "int", "%"],
            [23, "Unknown", "int", ""],
        ],
        "test_responses": [b"^D0941750,0560,0540,0000,060,0530,0420,0480,0480,0540,0,,,0,0480,000,0100,0175,010,020,020,080,0mr\r"],
    },
}
SETTER_COMMANDS = {}
COMMANDS_TO_REMOVE = []


#


class pi17m058(pi17):
    def __str__(self):
        return "PI17 protocol handler modified for model 058"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"PI17m058"
        # Add pi17m058 specific commands to pi17 commands
        self.COMMANDS.update(QUERY_COMMANDS)
        # Add pi17m058 specific setter commands
        self.COMMANDS.update(SETTER_COMMANDS)
        # remove and unwanted pi30 commands
        for item in COMMANDS_TO_REMOVE:
            self.COMMANDS.pop(item, None)



import abc
import calendar  # noqa: F401
import logging
import re
from datetime import datetime
from typing import Tuple
from pydantic import BaseModel
# from powermon.dto.protocolDTO import ProtocolDTO

from ..helpers import get_resp_defn, get_value
from .protocol_helpers import BigHex2Short, BigHex2Float  # noqa: F401
from .protocol_helpers import LittleHex2Float, LittleHex2Short  # noqa: F401
from .protocol_helpers import LittleHex2UInt, LittleHex2Int  # noqa: F401
from .protocol_helpers import Hex2Ascii, Hex2Int, Hex2Str  # noqa: F401
from .protocol_helpers import uptime  # noqa: F401
from .protocol_helpers import crcPI as crc

log = logging.getLogger("AbstractProtocol")


class ProtocolDTO(BaseModel):
    protocol_id: str
    commands: dict


class AbstractProtocol(metaclass=abc.ABCMeta):
    def __init__(self, *args, **kwargs) -> None:
        self._command = None
        self._command_dict = None
        self.COMMANDS = {}
        self.STATUS_COMMANDS = None
        self.SETTINGS_COMMANDS = None
        self.DEFAULT_COMMAND = None
        self.PID = None
        self.ID_COMMANDS = None
        self._protocol_id = None

    def toDTO(self) -> ProtocolDTO:
        dto = ProtocolDTO(protocol_id=self._protocol_id, commands=self.list_commands())
        return dto

    def list_commands(self) -> dict:
        # print(f"{'Parameter':<30}\t{'Value':<15} Unit")
        if self._protocol_id is None:
            log.error("Attempted to list commands with no protocol defined")
            return {"ERROR": ["Attempted to list commands with no protocol defined", ""]}
        result = {}
        result["_command"] = "command help"
        result["_command_description"] = f"List available commands for protocol {str(self._protocol_id, 'utf-8')}"
        for command in sorted(self.COMMANDS):
            if "help" in self.COMMANDS[command]:
                info = self.COMMANDS[command]["description"] + self.COMMANDS[command]["help"]
            else:
                info = self.COMMANDS[command]["description"]
            result[command] = [info, ""]
        return result

    def get_protocol_id(self) -> bytes:
        return self._protocol_id

    def get_full_command(self, command) -> bytes:
        log.info(f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands")
        byte_cmd = bytes(command, "utf-8")
        # calculate the CRC
        crc_high, crc_low = crc(byte_cmd)
        # combine byte_cmd, CRC , return
        full_command = byte_cmd + bytes([crc_high, crc_low, 13])
        log.debug(f"full command: {full_command}")
        return full_command

    def get_command_defn(self, command) -> dict:
        log.debug(f"Processing command '{command}'")
        if command in self.COMMANDS and "regex" not in self.COMMANDS[command]:
            log.debug(f"Found command {command} in protocol {self._protocol_id}")
            return self.COMMANDS[command]
        for _command in self.COMMANDS:
            if "regex" in self.COMMANDS[_command] and self.COMMANDS[_command]["regex"]:
                log.debug(f"Regex commands _command: {_command}")
                _re = re.compile(self.COMMANDS[_command]["regex"])
                match = _re.match(command)
                if match:
                    log.debug(f"Matched: {command} to: {self.COMMANDS[_command]['name']} value: {match.group(1)}")
                    self._command_value = match.group(1)
                    return self.COMMANDS[_command]
        log.info(f"No command_defn found for {command}")
        return None

    def get_responses(self, response) -> list:
        """
        Default implementation of split and trim
        """
        # Trim leading '(' + trailing CRC and \r of response, then split
        if type(response) is str:
            return response[1:-3].split(" ")
        return response[1:-3].split(b" ")

    def check_response_valid(self, response) -> Tuple[bool, dict]:
        """
        Simplest validity check, CRC checks should be added to individual protocols
        """
        if response is None:
            return False, {"validity check": ["Error: Response was empty", ""]}
        return True, {}

    def process_response(
        self,
        data_name=None,
        data_type=None,
        data_units=None,
        raw_value=None,
        frame_number=0,
        extra_info=None,
    ):
        template = None
        # Check for a format modifying template
        if ":" in data_type:
            data_type, template = data_type.split(":", 1)
            log.debug(f"Got template {template} for {data_name} {raw_value}")
        log.debug(f"Processing data_type: {data_type} for data_name: {data_name}, raw_value {raw_value}")
        if data_type == "loop":
            log.warning("loop not implemented...")
            return [(data_name, None, data_units, extra_info)]
        if data_type == "exclude" or data_type == "discard" or raw_value == "extra":
            # Just ignore these ones
            log.debug(f"Discarding {data_name}:{raw_value}")
            return [(None, raw_value, data_units, extra_info)]
        if data_type == "option":
            try:
                key = int(raw_value)
                r = data_units[key]
            except ValueError:
                r = f"Unable to process to int: {raw_value}"
                return [(None, r, "", None)]
            except IndexError:
                r = f"Invalid option: {key}"
            return [(data_name, r, "", extra_info)]
        if data_type == "hex_option":
            key = int(raw_value[0])
            if key < len(data_units):
                r = data_units[key]
            else:
                r = f"Invalid hex_option: {key}"
            return [(data_name, r, "", extra_info)]
        if data_type == "flags":
            log.debug("flags defn")
            # [
            #     "flags",
            #     "Device Status",
            #     [
            #         "Is SBU Priority Version Added",
            #         "Is SCC Firmware Updated",
            #         "Is Load On",
            #     ],
            # ],
            return_value = []
            for i, flag in enumerate(raw_value):
                return_value.append((data_units[i], int(chr(flag)), "bool", None))

            # if flag != "" and flag != b'':
            # msgs[resp_format[2][j]] = [int(flag), "bool"]
            # print(j, int(flag))
            return return_value
        if data_type == "keyed":
            log.debug("keyed defn")
            # [
            #     "keyed",
            #     1,
            #     "Command response flag",
            #     {
            #         "00": "OK",
            #         "01": "Unknown ID",
            #         "02": "Not supported",
            #         "04": "Parameter Error",
            #     },
            # ],
            key = ""
            for x in raw_value:
                key += f"{x:02x}"
            r = data_units.get(key, f"Invalid key: {key}")
            return [(data_name, r, "", None)]
        if data_type == "str_keyed":
            log.debug("str_keyed defn")
            # [
            #     "str_keyed",
            #     "Device Mode",
            #     {
            #         "B": "Inverter (Battery) Mode",
            #         "C": "PV charging Mode",
            #         "D": "Shutdown Mode",
            #         "F": "Fault Mode",
            #         "G": "Grid Mode",
            #         "L": "Line Mode",
            #         "P": "Power on Mode",
            #         "S": "Standby Mode",
            #         "Y": "Bypass Mode",
            #     },
            # ]
            key = raw_value.decode()
            r = data_units.get(key, f"Invalid key: {key}")
            return [(data_name, r, "", extra_info)]
        format_string = f"{data_type}(raw_value)"
        log.debug(f"Processing format string {format_string}")
        try:
            r = eval(format_string)
        except ValueError as e:
            log.info(f"Failed to eval format {format_string} (returning 0), error: {e}")
            return [(data_name, 0, data_units, extra_info)]
        except TypeError as e:
            log.warning(f"Failed to eval format {format_string}, error: {e}")
            return [(data_name, format_string, data_units, extra_info)]
        if template is not None:
            # eg template=r/1000
            r = eval(template)
        if "{" in data_name:
            # eg "f'Frame Number {f:02d}'"
            f = frame_number  # noqa: F841
            data_name = eval(data_name)
        return [(data_name, r, data_units, extra_info)]

    def decode_result(self, result, command):
        log.info("decode_result: raw: %s, command: %s" % (result.raw_response, command.name))

        # TODO: sort this so it isnt so carp
        data = self.decode(result.raw_response, command.name)
        # Clean data
        data.pop("raw_response", None)
        data.pop("_command", None)
        data.pop("_command_description", None)
        result.decoded_response = data
        return result

    def decode(self, response, command) -> dict:
        """
        Take the raw response and turn it into a dict of name: value, unit entries
        """

        log.info(f"response passed to decode: {response}")
        msgs = {}

        # Add metadata
        msgs["_command"] = command
        # Check for a stored command definition
        command_defn = self.get_command_defn(command)
        if command_defn is not None:
            msgs["_command_description"] = command_defn["description"]
            len_command_defn = len(command_defn["response"])

        # Check response is valid
        valid, _msg = self.check_response_valid(response)
        if not valid:
            msgs.update(_msg)
            log.info(f"validity check fail: {_msg}")
            return msgs

        # Add Raw response
        _response = b""
        for item in response:
            if type(item) is int:
                _response += chr(item).encode()
            else:
                _response += item.encode()
        raw_response = _response.decode("utf-8")
        msgs["raw_response"] = [raw_response, ""]

        if command_defn is None:
            # No definition, so just return the data
            len_command_defn = 0
            log.debug(f"No definition for command {command}, (splitted) raw response returned")
            msgs["WARNING"] = [
                f"No definition for command {command} in protocol {self._protocol_id}",
                "",
            ]
            msgs["response"] = [raw_response, ""]
            return msgs

        # Determine the type of response
        if "response_type" in command_defn:
            response_type = command_defn["response_type"]
        else:
            response_type = "DEFAULT"
        log.info(f"Processing response of type {response_type}")

        # Split the response into individual responses
        responses = self.get_responses(response)
        log.debug(f"trimmed and split responses: {responses}")

        # Decode response based on stored command definition and type
        # process default response type
        # TODO: fix this - move into new approach
        # DEFAULT - responses are determined by the order they are returned
        if response_type == "DEFAULT":
            log.info("Processing DEFAULT type responses")
            # print("Processing DEFAULT type responses")
            for i, result in enumerate(responses):
                # decode result
                if result == b'':
                    continue
                if type(result) is bytes:
                    result = result.decode("utf-8")

                # Check if we are past the 'known' responses
                if i >= len_command_defn:
                    resp_format = ["string", f"Unknown value in response {i}", ""]
                else:
                    resp_format = command_defn["response"][i]

                # key = "{}".format(resp_format[1]).lower().replace(" ", "_")
                key = resp_format[1]
                log.debug(f'result {result}, key {key}, resp_format {resp_format}')
                # Process results
                if result == "NAK":
                    msgs[f"WARNING{i}"] = [
                        f"Command {command} was rejected",
                        "",
                    ]
                elif resp_format[0] == "float":
                    try:
                        result = float(result)
                    except ValueError:
                        log.debug(f"Error resolving {result} as float")
                    msgs[key] = [result, resp_format[2]]
                elif resp_format[0] == "int":
                    try:
                        result = int(result)
                    except ValueError:
                        log.debug(f"Error resolving {result} as int")
                    msgs[key] = [result, resp_format[2]]
                elif resp_format[0] == "string":
                    msgs[key] = [result, resp_format[2]]
                elif resp_format[0] == "10int":
                    if "--" in result:
                        result = 0
                    msgs[key] = [float(result) / 10, resp_format[2]]
                # eg. ['option', 'Output source priority', ['Utility first', 'Solar first', 'SBU first']],
                elif resp_format[0] == "option":
                    msgs[key] = [resp_format[2][int(result)], ""]
                # eg. ['keyed', 'Machine type', {'00': 'Grid tie', '01': 'Off Grid', '10': 'Hybrid'}],
                elif resp_format[0] == "keyed":
                    msgs[key] = [resp_format[2][result], ""]
                # eg. ['flags', 'Device status', [ 'is_load_on', 'is_charging_on' ...
                elif resp_format[0] == "flags":
                    for j, flag in enumerate(result):
                        # if flag != "" and flag != b'':
                        msgs[resp_format[2][j]] = [int(flag), "bool"]
                # eg. ['stat_flags', 'Warning status', ['Reserved', 'Inver...
                elif resp_format[0] == "stat_flags":
                    output = ""
                    for j, flag in enumerate(result):
                        # only display 'enabled' flags
                        # if flag == "1" or flag == b"1":
                        #    output = "{}\n\t- {}".format(output, resp_format[2][j])
                        # display all flags
                        key = resp_format[2][j]
                        output = flag
                        if key:  # only add msg if key is something
                            msgs[key] = [output, ""]
                # eg. ['enflags', 'Device Status', {'a': {'name': 'Buzzer', 'state': 'disabled'},
                elif resp_format[0] == "enflags":
                    # output = {}
                    status = "unknown"
                    for item in result:
                        if item == "E":
                            status = "enabled"
                        elif item == "D":
                            status = "disabled"
                        else:
                            # output[resp_format[2][item]['name']] = status
                            # _key = "{}".format(resp_format[2][item]["name"]).lower().replace(" ", "_")
                            if resp_format[2].get(item, None):
                                _key = resp_format[2][item]["name"]
                            else:
                                _key = "unknown_{}".format(item)
                            msgs[_key] = [status, ""]
                    # msgs[key] = [output, '']
                elif resp_format[0] == "multi":
                    for x, item in enumerate(result):
                        item_value = int(item)
                        item_resp_format = resp_format[1][x]
                        item_type = item_resp_format[0]
                        # print(x, item_value, item_resp_format, item_type)
                        if item_type == "option":
                            item_name = item_resp_format[1]
                            resolved_value = item_resp_format[2][item_value]
                            msgs[item_name] = [resolved_value, ""]
                        elif item_type == "string":
                            item_name = item_resp_format[1]
                            msgs[item_name] = [item_value, ""]
                        else:
                            print(f"item type {item_type} not defined")
                elif command_defn["type"] == "SETTER":
                    # _key = "{}".format(command_defn["name"]).lower().replace(" ", "_")
                    _key = command_defn["name"]
                    msgs[_key] = [result, ""]
                else:
                    log.info(f"Processing unknown response format {result}")
                    msgs[i] = [result, ""]
            return msgs

        # Check for multiple frame type responses
        if response_type == "MULTIFRAME-POSITIONAL":
            log.debug("Processing MULTIFRAME-POSITIONAL type responses")
            # MULTIFRAME-POSITIONAL - multiple frames of responses are not separated and are determined by the position in the response
            # each frame has the same definition
            frame_count = len(responses)
            log.debug(f"got {frame_count} frames")
            # the responses are the frames
            frames = responses
        else:
            frames = [responses]
            frame_count = 1

        for frame_number, frame in enumerate(frames):
            for i, response in enumerate(frame):
                extra_info = None
                if response_type == "KEYED":
                    log.debug("Processing KEYED type responses")
                    # example defn ["V", "Main or channel 1 (battery) voltage", "V", "float:r/1000"]
                    # example response data [b'H1', b'-32914']
                    if len(response) <= 1:
                        # Not enough data in response, so ignore
                        continue
                    lookup_key = response[0]
                    raw_value = response[1]
                    response_defn = get_resp_defn(lookup_key, command_defn["response"])
                    if response_defn is None:
                        # No definition for this key, so ignore???
                        log.warning(f"No definition for {response}")
                        continue
                    # key = response_defn[0] #0
                    data_type = response_defn[3]  # 1
                    data_name = response_defn[1]  # 2
                    data_units = response_defn[2]  # 3

                elif response_type == "SEQUENTIAL":
                    log.debug("Processing SEQUENTIAL type responses")
                    # check for extra definitions...
                    extra_responses_needed = len(command_defn["response"]) - len(frame)
                    if extra_responses_needed > 0:
                        for _ in range(extra_responses_needed):
                            frame.append("extra")
                    # example ["int", "Energy produced", "Wh"]

                    # Check if we are past the 'known' responses
                    if i >= len_command_defn:
                        response_defn = ["str", f"Unknown value in response {i}", ""]
                    else:
                        response_defn = command_defn["response"][i]
                    log.debug(f"Got defn {response_defn}")
                    raw_value = response
                    # spacer = response_defn[0] #0
                    data_type = response_defn[0]  # 1
                    data_name = response_defn[1]  # 2
                    data_units = response_defn[2]  # 3

                elif response_type == "INDEXED":
                    log.debug("Processing INDEXED type responses")
                    # [1, "AC Input Voltage", "float", "V", {icon: blah}]
                    # check for extra definitions...
                    extra_responses_needed = len(command_defn["response"]) - len(frame)
                    if extra_responses_needed > 0:
                        for _ in range(extra_responses_needed):
                            frame.append("extra")

                    # Check if we are past the 'known' responses
                    if i >= len_command_defn:
                        if not response:
                            continue
                        response_defn = [
                            i + 1,
                            f"Unknown value in response {i+1}",
                            "str",
                            "",
                        ]
                    else:
                        response_defn = command_defn["response"][i]
                    log.debug(f"Got defn {response_defn}")
                    raw_value = response
                    # data_posi = get_value(response_defn, 0)
                    data_name = get_value(response_defn, 1)
                    data_type = get_value(response_defn, 2)
                    data_units = get_value(response_defn, 3)
                    extra_info = get_value(response_defn, 4)

                    # print(f"{data_type=}, {data_name=}, {raw_value=}")
                elif response_type in ["POSITIONAL", "MULTIFRAME-POSITIONAL"]:
                    log.debug("Processing POSITIONAL type responses")
                    # check for extra definitions...
                    extra_responses_needed = len(command_defn["response"]) - len(frame)
                    if extra_responses_needed > 0:
                        for _ in range(extra_responses_needed):
                            frame.append("extra")
                    # POSITIONAL - responses are not separated and are determined by the position in the response
                    # example defn :
                    #   ["discard", 1, "start flag", ""],
                    #   ["BigHex2Short", 2, "Battery Bank Voltage", "V"],
                    # example response data:
                    #   [b'\xa5', b'\x01', b'\x90', b'\x08', b'\x01\t', b'\x00\x00', b'u\xcf', b'\x03\x99', b'']
                    raw_value = response
                    # Check if we are past the 'known' responses
                    if i >= len_command_defn:
                        response_defn = ["str", 1, f"Unknown value in response {i}", ""]
                    else:
                        response_defn = command_defn["response"][i]
                    if response_defn is None:
                        # No definition for this key, so ignore???
                        log.warning(f"No definition for {response}")
                        response_defn = [
                            "str",
                            1,
                            f"Undefined value in response {i}",
                            "",
                        ]
                    log.debug(f"Got defn {response_defn}")
                    # length = response_defn[1] #0
                    data_type = response_defn[0]  # 1
                    data_name = response_defn[2]  # 2
                    data_units = response_defn[3]  # 3

                # Check for lookup
                if data_type.startswith("lookup"):
                    log.debug("processing lookup...")
                    log.info(f"Processing data_type: '{data_type}' for data_name: '{data_name}', raw_value '{raw_value}'")
                    m = msgs
                    template = data_type.split(":", 1)[1]
                    log.debug(f"Got template {template} for {data_name} {raw_value}")
                    lookup = eval(template)
                    log.debug(f"looking up values for: {lookup}")
                    value, data_units = m[lookup]
                    if data_name is not None:
                        msgs[data_name] = [value, data_units, extra_info]
                elif data_type.startswith("info"):
                    log.debug("processing info...")
                    # print(
                    #    f"Processing {data_type=} for {data_name=}, {data_units=} {response=} {command=} {self._command_value=}"
                    # )
                    template = data_type.split(":", 1)[1]
                    # Provide cv as shortcut to self._command_value for info fields
                    cv = self._command_value  # noqa: F841
                    value = eval(template)
                    if data_name is not None:
                        msgs[data_name] = [value, data_units, extra_info]
                else:
                    # Process response
                    processed_responses = self.process_response(
                        data_name=data_name,
                        raw_value=raw_value,
                        data_units=data_units,
                        data_type=data_type,
                        frame_number=frame_number,
                        extra_info=extra_info,
                    )
                    # data_name, value, data_units = self.process_response(
                    #     data_name=data_name,
                    #     raw_value=raw_value,
                    #     data_units=data_units,
                    #     data_type=data_type,
                    #     frame_number=frame_number,
                    # )
                    # print(data_type, data_name, raw_value, value)
                    for item in processed_responses:
                        data_name, value, data_units, extra_info = item
                        if data_name is not None:
                            if extra_info:
                                msgs[data_name] = [value, data_units, extra_info]
                            else:
                                msgs[data_name] = [value, data_units]
            # print(f"{i=} {response=} {len(command_defn['response'])}")

        return msgs



from sqlalchemy.orm import Session

from . import models, schemas


def get_user(db: Session, user_id: int):
    return db.query(models.User).filter(models.User.id == user_id).first()


def get_user_by_email(db: Session, email: str):
    return db.query(models.User).filter(models.User.email == email).first()


def get_users(db: Session, skip: int = 0, limit: int = 100):
    return db.query(models.User).offset(skip).limit(limit).all()


def create_user(db: Session, user: schemas.UserCreate):
    fake_hashed_password = user.password + "notreallyhashed"
    db_user = models.User(email=user.email, hashed_password=fake_hashed_password)
    db.add(db_user)
    db.commit()
    db.refresh(db_user)
    return db_user


def get_items(db: Session, skip: int = 0, limit: int = 100):
    return db.query(models.Item).offset(skip).limit(limit).all()


def create_user_item(db: Session, item: schemas.ItemCreate, user_id: int):
    db_item = models.Item(**item.dict(), owner_id=user_id)
    db.add(db_item)
    db.commit()
    db.refresh(db_item)
    return db_item


# New stuff
def create_mqtt_message(db: Session, mqtt_message: schemas.MQTTMessageCreate):
    db_mqtt_message = models.MQTTMessage(topic=mqtt_message.topic, message=mqtt_message.message)
    db.add(db_mqtt_message)
    db.commit()
    db.refresh(db_mqtt_message)
    return db_mqtt_message

def get_mqtt_messages(db: Session, skip: int = 0, limit: int = 100):
    return db.query(models.MQTTMessage).offset(skip).limit(limit).all()

def get_mqtt_message(db: Session, mqtt_message_id: int):
    return db.query(models.MQTTMessage).filter(models.MQTTMessage.id == mqtt_message_id).first()


import threading
import time
from time import sleep


class RepeatedTimer(object):
    def __init__(self, interval, function, *args, **kwargs):
        self._timer = None
        self.interval = interval
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.is_running = False
        self.next_call = time.time()
        self.start()

    def _run(self):
        self.is_running = False
        self.start()
        self.function(*self.args, **self.kwargs)

    def start(self):
        if not self.is_running:
            self.next_call += self.interval
            self._timer = threading.Timer(self.next_call - time.time(), self._run)
            self._timer.start()
            self.is_running = True

    def stop(self):
        self._timer.cancel()
        self.is_running = False


if __name__ == "__main__":
    # main()
    def hello(name):
        global lock
        print("Hello %s! lock=%s" % (name, lock))
        lock = not lock
    lock = False
    # some_lock = threading.Lock()
    # with some_lock:
    #     # do something...
    # is equivalent to:
    # some_lock.acquire()
    # try:
    #     # do something...
    # finally:
    #     some_lock.release()
    print("starting...")
    rt = RepeatedTimer(2, hello, "World")  # it auto-starts, no need of rt.start()
    rt2 = RepeatedTimer(3, hello, "Bob")
    try:
        sleep(20)  # your long-running job goes here...
    finally:
        rt.stop()  # better in a try/finally block to make sure the program ends!
        rt2.stop()



import logging

from .jkabstractprotocol import jkAbstractProtocol


log = logging.getLogger("jk04")

NEW_COMMANDS = {
    "getCellData": {
        "name": "getCellData",
        "command_code": "96",
        "record_type": "2",
        "description": "BLE Cell Data inquiry",
        "help": " -- queries the ble device for the cell data",
        "type": "QUERY",
        "response_type": "POSITIONAL",
        "response": [
            ["Hex2Str", 4, "Header", ""],
            ["Hex2Str", 1, "Record Type", ""],
            ["Hex2Int", 1, "Record Counter", ""],
            ["LittleHex2Float", 4, "Voltage Cell01", "V"],
            ["LittleHex2Float", 4, "Voltage Cell02", "V"],
            ["LittleHex2Float", 4, "Voltage Cell03", "V"],
            ["LittleHex2Float", 4, "Voltage Cell04", "V"],
            ["LittleHex2Float", 4, "Voltage Cell05", "V"],
            ["LittleHex2Float", 4, "Voltage Cell06", "V"],
            ["LittleHex2Float", 4, "Voltage Cell07", "V"],
            ["LittleHex2Float", 4, "Voltage Cell08", "V"],
            ["LittleHex2Float", 4, "Voltage Cell09", "V"],
            ["LittleHex2Float", 4, "Voltage Cell10", "V"],
            ["LittleHex2Float", 4, "Voltage Cell11", "V"],
            ["LittleHex2Float", 4, "Voltage Cell12", "V"],
            ["LittleHex2Float", 4, "Voltage Cell13", "V"],
            ["LittleHex2Float", 4, "Voltage Cell14", "V"],
            ["LittleHex2Float", 4, "Voltage Cell15", "V"],
            ["LittleHex2Float", 4, "Voltage Cell16", "V"],
            ["LittleHex2Float", 4, "Voltage Cell17", "V"],
            ["LittleHex2Float", 4, "Voltage Cell18", "V"],
            ["LittleHex2Float", 4, "Voltage Cell19", "V"],
            ["LittleHex2Float", 4, "Voltage Cell20", "V"],
            ["LittleHex2Float", 4, "Voltage Cell21", "V"],
            ["LittleHex2Float", 4, "Voltage Cell22", "V"],
            ["LittleHex2Float", 4, "Voltage Cell23", "V"],
            ["LittleHex2Float", 4, "Voltage Cell24", "V"],
            ["LittleHex2Float", 4, "Resistance Cell01", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell02", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell03", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell04", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell05", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell06", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell07", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell08", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell09", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell10", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell11", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell12", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell13", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell14", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell15", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell16", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell17", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell18", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell19", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell20", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell21", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell22", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell23", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell24", "Ohm"],
            ["LittleHex2Float", 4, "Resistance Cell25", "Ohm"],
            ["LittleHex2Float", 4, "Average Cell Voltage", "V"],
            ["LittleHex2Float", 4, "Delta Cell Voltage", "V"],
            ["discard", 4, "", ""],
            ["discard", 4, "", ""],
            ["Hex2Int:r+1", 1, "Highest Cell", ""],
            ["Hex2Int:r+1", 1, "Lowest Cell", ""],
            ["Hex2Str", 2, "Flags", ""],
            ["discard", 4, "", ""],
            ["discard", 7, "", ""],
            ["discard", 4, "", ""],
            ["discard", 4, "", ""],
            ["discard", 45, "", ""],
            ["uptime", 3, "uptime", ""],
            ["discard", 5, "", ""],
            ["discard", 4, "", ""],
            ["discard", 1, "", ""],
            ["Hex2Str", 1, "Checksum", ""],
            [
                "lookup:'Voltage Cell{:02d}'.format(m['Highest Cell'][0])",
                0,
                "Highest Cell Voltage",
                "V",
            ],
            [
                "lookup:'Voltage Cell{:02d}'.format(m['Lowest Cell'][0])",
                0,
                "Lowest Cell Voltage",
                "V",
            ],
        ],
        "test_responses": [
            bytes.fromhex(
                "55aaeb9002ff5b566240e34e62406e6a62404a506240acd7624011d26240bddd62409ad1624044c86240cedc6240ccc7624079e1624057dc624073a262405f80624088c46240000000000000000000000000000000000000000000000000000000000000000013315c3d0636143d26e0113d8021f03c1153363d8980123d7e7c033dac41233d1ad83c3d9d6f4f3d8eb51e3d6a2c293deb28653d189c523da3724e3deb94493d9ab2c23d00000000000000000000000000000000000000000000000000000000000000001aad62400084053c00000000ffff00000b000000000000000000000000000036a3554c40000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000000be0b54001456a43fb876a43f00a2"
            ),
            b"U\xaa\xeb\x90\x02\xfd\x01\x04\x13@\x81\xbc\x16@E\xd2\x10@\xed\xd4\x16@\xed\xd4\x16@2\x1e\x17@\xa8\x10\x14@\xe3\x7f\x17@\x15\xa4\x16@\xf7)\x16@2\x1e\x17@\xb1\xf4\x0b@2\xa3\x14@\x9eJ\r@\x9e\xc5\x0f@\xa8\x8b\x16@\x9e6\x17@\xc6\x05\x17@\xe3\x7f\x17@Y\xed\x16@\xe3\x7f\x17@\xcf\xdf\x13@Y\xed\x16@2\xa3\x14@\xab\xe5p>Yk2>&\xef\xf6=>\xb84>p\xfc~>\xab9\xbc>\xde\xd3\xb6>25\x80>672>\xaeG\xf7=\x86\xc4\xfa=g,\x02>\xf6&\x02>\x97S\x01>\xd8\x1d\x01>\x94%\x05>JF\x00>\x8f\xd83>\xe0a\x92>\x05\xf2\xaa>\xd2\xbaU>\xad\xc0\xf8=\xee\x88\xf7=\xd5\xa2@>\x00\x00\x00\x00\x92\xf2\x14@P,7>\x00\x00\x00\x00\xff\xff\xff\x00\x07\x0b\x01\x01\x00X\xb6?\x00\x00\x00\x00\x00\x00\x00Z{\xedK@\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\x00\x00\xd2\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xa0/\x00\x00\x00\x00\x00\x00\x00X*@\x00\x0b",
            b"U\xaa\xeb\x90\x02\x12O\xa2\x12@\xa8\x8b\x16@\x94p\x10@\xed\xd4\x16@<s\x16@\xed\xd4\x16@\xf7\xae\x13@\xb2\xe0\x15@\x0bO\x17@wg\x17@\x0bO\x17@\x00\x93\x0b@Yr\x14@\xec\xe8\x0c@\xedc\x0f@2\x1e\x17@\xc6\x05\x17@\x81\xbc\x16@\xbc\xb0\x17@\xa8\x8b\x16@\xbc\xb0\x17@\x8a\x96\x13@\x81\xbc\x16@Yr\x14@\xab\xe5p>Yk2>&\xef\xf6=>\xb84>p\xfc~>\xab9\xbc>\xde\xd3\xb6>25\x80>672>\xaeG\xf7=\x86\xc4\xfa=g,\x02>\xf6&\x02>\x97S\x01>\xd8\x1d\x01>\x94%\x05>JF\x00>\x8f\xd83>\xe0a\x92>\x05\xf2\xaa>\xd2\xbaU>\xad\xc0\xf8=\xee\x88\xf7=\xd5\xa2@>\x00\x00\x00\x00\n\xd4\x14@@\xce>>\x00\x00\x00\x00\xff\xff\xff\x00\x07\x0b\x02\x01\x004\x00\xc0\x00\x00\x00\x00\x00\x00\x00\xf6\xc1\xe7K@\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x01\x01\xc9\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\xd08\x00\x00\x00\x00\x00\x00\x00(*@\x00\xea",
        ],
    },
}


class jk04(jkAbstractProtocol):
    """
    JK04 - Handler for JKBMS 4 byte data communication
         - e.g. 5b566240 = 3.5365V
    """

    def __str__(self):
        return "JK04 - JKBMS BLE 4 byte data communication protocol handler"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"JK04"
        self.COMMANDS.update(NEW_COMMANDS)
        self.STATUS_COMMANDS = [
            "getCellData",
        ]
        self.SETTINGS_COMMANDS = [
            "getInfo",
        ]
        self.DEFAULT_COMMAND = "getCellData"



import logging

from .abstractprotocol import AbstractProtocol
from .protocol_helpers import crcJK232 as crc


log = logging.getLogger("jk232")


# Read basic information and status
# DD A5 03 00 FF FD 77
# start bit  0xDD
# status 0xA5 means read, status 0x5A means write.
# command code 0x03
# Data length: 1 byte, indicating the effective length of the data carried in the frame.
# Data content: N bytes, the content carried by the frame data, when the data length is 0, there is no such part.
# Verification: 2 bytes,
# the verification field is "command code + length byte + data segment content",
#    the verification method is thesum of the above fields and then the inverse plus 1, the high bit is in the front and the low bit is in the back.
# Stop bit: 1 byte, indicating the end of a frame of data, fixed as 0x77;


COMMANDS = {
    "getBalancerData": {
        "name": "getBalancerData",
        "command_code": "03",
        "description": "Get Balancer Data",
        "help": " -- Get Balancer Data",
        "type": "QUERY",
        "checksum_required": "True",
        "response_type": "POSITIONAL",
        "response": [
            ["Hex2Str", 1, "Start Byte", ""],
            ["Hex2Str", 1, "Command Code", ""],
            ["Hex2Str", 1, "Status", ""],
            ["Hex2Int", 1, "Data Length", ""],
            ["BigHex2Short:r/100", 2, "Total Battery Voltage", "V"],
            ["BigHex2Short:r/100", 2, "Total Current", "A"],
            ["BigHex2Short:r/100", 2, "Remaining Capacity", "Ah"],
            ["BigHex2Short:r/100", 2, "Nominal Capacity", "Ah"],
            ["BigHex2Short", 2, "Cycles", "cycles"],
            ["Hex2Str", 2, "Production Date", ""],
            ["Hex2Str", 2, "Equilibrium State (TODO)", ""],
            ["Hex2Str", 2, "Equilibrium State 2 (TODO)", ""],
            ["Hex2Str", 2, "Protection State (TODO)", ""],
            ["Hex2Str", 1, "Keep", ""],
            ["Hex2Int", 1, "Remaining Battery", "%"],
            ["Hex2Str", 1, "FET Control Status", ""],
            ["Hex2Int", 1, "Number of Battery Strings", ""],
            ["Hex2Int", 1, "Number of NTC", ""],
            ["BigHex2Short:(r-2731)/10", 2, "NTC 1", "°C"],
            ["BigHex2Short:(r-2731)/10", 2, "NTC 2", "°C"],
            ["Hex2Str", 2, "Checksum", ""],
            ["Hex2Str", 1, "End Byte", ""],
        ],
        "test_responses": [
            bytes.fromhex(
                "DD 03 00 1B 17 00 00 00 02 D0 03 E8 00 00 20 78 00 00 00 00 00 00 10 48 03 0F 02 0B 76 0B 82 FB FF 77"
            ),
        ],
    },
}


class jk232(AbstractProtocol):
    def __str__(self):
        return "JKBMS RS232 serial communication protocol handler"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"JK232"
        self.COMMANDS = COMMANDS
        self.STATUS_COMMANDS = [
            "getBalancerData",
        ]
        self.SETTINGS_COMMANDS = [
            "",
        ]
        self.DEFAULT_COMMAND = "getBalancerData"

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different
        """
        log.info(f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands")
        # These need to be set to allow other functions to work`
        self._command = command
        self._command_defn = self.get_command_defn(command)
        # End of required variables setting
        if self._command_defn is None:
            # Maybe return a default here?
            return None
        if "command_code" in self._command_defn:

            # Read basic information and status
            # DD A5 03 00 FF FD 77
            # full command is 7 bytes long
            cmd = bytearray(7)

            # start bit  0xDD
            cmd[0] = 0xDD
            log.debug(f"cmd with start bit: {cmd}")

            # status 0xA5 means read, status 0x5A means write.
            if self._command_defn["type"] == "SETTER":
                cmd[1] = 0x5A
            else:
                cmd[1] = 0xA5
            # command code 0x03
            command_code = int(self._command_defn["command_code"], 16)
            # Data length: 1 byte, indicating the effective length of the data carried in the frame.
            # Data content: N bytes, the content carried by the frame data, when the data length is 0, there is no such part.
            data = ""
            # TODO: data stuff here
            data_len = len(data)
            if data_len == 0:
                crc_high, crc_low = crc([command_code, data_len])
                cmd[2] = command_code
                cmd[3] = data_len
                cmd[4] = crc_high
                cmd[5] = crc_low
                cmd[6] = 0x77

            log.debug(f"cmd with crc: {cmd}")
            return cmd

    def get_responses(self, response):
        """
        Override the default get_responses as its different
        """
        responses = []
        # remove \n
        # response = response.replace(b"\n", b"")

        if self._command_defn is not None and self._command_defn["response_type"] == "POSITIONAL":
            # Have a POSITIONAL type response, so need to break it up...
            # example defn :
            # "response": [
            #   ["discard", 1, "start flag", ""],
            #   ["discard", 1, "module address", ""],
            #   ["discard", 1, "command id", ""],
            #   ["discard", 1, "data length", ""],
            # ]
            # example response data b"\xa5\x01\x90\x08\x02\x10\x00\x00uo\x03\xbc\xf3",
            for defn in self._command_defn["response"]:
                size = defn[1]
                item = response[:size]
                responses.append(item)
                response = response[size:]
            if response:
                responses.append(response)
            log.debug(f"get_responses: responses {responses}")
            return responses
        else:
            return bytearray(response)



""" formats / simple.py """
import logging
from powermon.formats.abstractformat import AbstractFormat
from powermon.dto.formatDTO import FormatDTO
from powermon.commands.result import Result
from powermon.commands.reading import Reading

log = logging.getLogger("simple")


class SimpleFormat(AbstractFormat):
    def __init__(self, formatConfig):
        super().__init__(formatConfig)
        self.name = "simple"
        self.extra_info = formatConfig.get("extra_info", False)
   
    def set_command_description(self, command_description):
        pass

    def format(self, result: Result) -> list:

        _result = []

        # check for error in result
        #if result.error:
        #    data = {}
        #    data["Error"] = [f"Command: {result.command_code} incurred an error or errors during execution or processing", ""]
        #    data["Error Count"] = [len(result.error_messages), ""]
        #    for i, message in enumerate(result.error_messages):
        #        data[f"Error #{i}"] = [message, ""]


        if len(result.get_responses()) == 0:
            return _result

        display_data : list[Reading] = self.format_and_filter_data(result)

        # build data to display
        for response in display_data:
            name = response.get_data_name()
            value = response.get_data_value()
            unit = response.get_data_unit()
            if self.extra_info:
                extra = ""
                if response.get_device_class() is not None:
                    extra = " " + response.get_device_class()
                if response.get_icon() is not None:
                    extra += " " + response.get_icon()
                if response.get_state_class() is not None:
                    extra += " " + response.get_state_class()
                
                
                _result.append(f"{name}={value}{unit}{extra}")
            else:
                _result.append(f"{name}={value}{unit}")
        return _result

    @classmethod
    def from_DTO(cls, dto: FormatDTO):
        return cls(formatConfig={})


"""
mqtt output module
outputs messages to mqtt broker
"""
import logging

from powermon.commands.result import Result
# from powermon.device import Device
# from powermon.formats.abstractformat import AbstractFormat
# from powermon.dto.outputDTO import OutputDTO
from powermon.outputs.abstractoutput import AbstractOutput

log = logging.getLogger("MQTT")


class MQTT(AbstractOutput):
    """ mqtt output class"""
    def __init__(self, results_topic: str):
        super().__init__(name="Mqtt")
        self.results_topic = results_topic

    def __str__(self):
        return f"outputs.MQTT: {self.results_topic=}"

    def get_topic(self) -> str:
        return self.results_topic

    def process(self, result: Result, mqtt_broker=None):
        log.info("Using output processor: MQTT")
        # exit if no data
        if result is None:
            log.debug("No result to output")
            return

        # Not sure that formatter and mqtt_broker are set, could use builder pattern to ensure they are set
        if self.formatter is None:
            log.error("No formatter supplied")
            raise RuntimeError("No formatter supplied")

        if mqtt_broker is None:
            log.error("No mqtt broker supplied")
            raise RuntimeError("No mqtt broker supplied")

        # build the messages...
        formatted_data = self.formatter.format(result)
        log.debug("mqtt.output msgs %s", formatted_data)

        # publish
        # TODO: check this approach (single vs multiple and tidy/consolidate)
        if self.formatter.sendsMultipleMessages():
            mqtt_broker.publishMultiple(formatted_data)
        else:
            mqtt_broker.publish(self.results_topic, formatted_data)

    @classmethod
    def from_config(cls, output_config) -> "MQTT":
        """build object from config dict"""
        results_topic = output_config.get("topic_override", None)
        return cls(results_topic=results_topic)



import logging
import time
import datetime
from strenum import LowercaseStrEnum
from enum import auto
from powermon.dto.triggerDTO import TriggerDTO


log = logging.getLogger("Trigger")


class TriggerType(LowercaseStrEnum):
    EVERY = auto()
    LOOPS = auto()
    AT = auto()
    ONCE = auto()
    DISABLED = auto()


class Trigger:

    DATE_FORMAT = "%d %b %Y %H:%M:%S"

    def __init__(self, trigger_type, value=None):
        self.trigger_type = trigger_type
        self.value = value
        self.togo = 0
        self.last_run : float | None = None
        self.next_run : float = self.determine_next_run()

    def __str__(self):
        return f"trigger: {self.trigger_type} {self.value} loops togo: {self.togo}"

    @classmethod
    def fromConfig(cls, config=None):
        if not config:
            # no trigger defined, default to every 60 seconds
            trigger_type = TriggerType.EVERY
            value = 60
        elif TriggerType.EVERY in config:
            trigger_type = TriggerType.EVERY
            value = config.get(TriggerType.EVERY, 61)
        elif TriggerType.LOOPS in config:
            trigger_type = TriggerType.LOOPS
            value = config.get(TriggerType.LOOPS, 101)
        elif TriggerType.AT in config:
            trigger_type = TriggerType.AT
            value = config.get(TriggerType.AT, "12:01")
        elif TriggerType.ONCE in config:
            trigger_type = TriggerType.ONCE
            value = config.get(TriggerType.ONCE, 0)
        else:
            trigger_type = TriggerType.DISABLED
            value = None
        return cls(trigger_type=trigger_type, value=value)

    @classmethod
    def from_DTO(cls, dto: TriggerDTO) -> "Trigger":
        return cls(trigger_type=dto.trigger_type, value=dto.value)    

    def touch(self):
        # store run time (as secs since epoch)
        self.last_run = time.time()
        # update next run time
        self.next_run = self.determine_next_run()

    def get_last_run(self) -> str:
        last_run_str = "Not yet run"
        if self.last_run is not None:
            last_run_str = time.strftime(Trigger.DATE_FORMAT, time.localtime(self.last_run))
        return last_run_str

    def get_next_run(self) -> str:
        next_run_str = "unknown"
        if self.next_run is not None:
            next_run_str = time.strftime(Trigger.DATE_FORMAT, time.localtime(self.next_run))
        return next_run_str

    def is_due(self) -> bool:
        # Store the time now
        now = time.time()
        if self.trigger_type == TriggerType.DISABLED:
            return False
        elif self.trigger_type == TriggerType.EVERY:
            if self.last_run is None:
                return True  # if hasnt run, run now
            if self.next_run <= now:
                return True
            return False
        elif self.trigger_type == TriggerType.LOOPS:
            if self.togo <= 0:
                self.togo = self.value
                return True
            else:
                self.togo -= 1
                return False
        elif self.trigger_type == TriggerType.AT:
            if self.next_run is None:
                #log.warning("at type trigger failed to set next run for %s" % command)
                return False
            if self.next_run <= now:
                return True
            return False
        elif self.trigger_type == TriggerType.ONCE:
            if int(self.value) == 0:
                self.value = 1
                return True
            else:
                return False
        #log.warning("no isDue set for %s" % command)
        return False

    def determine_next_run(self) -> float:
        #TODO: split this into a function per trigger type
        if self.trigger_type == TriggerType.EVERY:
            # triggers every xx seconds
            # if hasnt run, run now
            if self.last_run is None:
                self.next_run = time.time()
            else:
                self.next_run = self.last_run + self.value
        elif self.trigger_type == TriggerType.AT:
            # triggers at specific time each day
            dt_today = datetime.datetime.now()
            dt_now = dt_today.time()
            at_time = datetime.time.fromisoformat(self.value)
            if dt_now < at_time:
                # needs to run today at at_time
                self.next_run = dt_today.replace(hour=at_time.hour, minute=at_time.minute, second=at_time.second, microsecond=0).timestamp()
            else:
                # needs to run tomorrow at at_time
                self.next_run = (dt_today.replace(hour=at_time.hour, minute=at_time.minute, second=at_time.second, microsecond=0) + datetime.timedelta(days=1)).timestamp()
        else:
            self.next_run = None
        return self.next_run

    def to_dto(self):
        return TriggerDTO(
            trigger_type=self.trigger_type,
            value=self.value
        )



import logging

from .jkabstractprotocol import jkAbstractProtocol
from .protocol_helpers import crc8


log = logging.getLogger("jk485")


# Request balancer data
# 55 AA 01 FF 00 00 FF
# frame header  0x55 0xaa
# slave address 0x01
# command code  0xff
# frame data    0x0000
# checksum      0xff

COMMANDS = {
    "getBalancerData": {
        "name": "getBalancerData",
        "command_code": "FF",
        "description": "Get Balancer Data",
        "help": " -- Get Balancer Data",
        "type": "QUERY",
        "checksum_required": "True",
        "response_type": "POSITIONAL",
        "response": [
            ["Hex2Str", 2, "Header", ""],
            ["Hex2Str", 1, "Slave Address", ""],
            ["Hex2Str", 1, "Command Code", ""],
            ["BigHex2Short:r/100", 2, "Total Battery Voltage", "V"],
            ["BigHex2Short:r/1000", 2, "Average Cell Voltage", "V"],
            ["Hex2Int", 1, "Number of Cells", ""],
            ["Hex2Int", 1, "Highest Cell", ""],
            ["Hex2Int", 1, "Lowest Cell", ""],
            ["Hex2Str", 1, "Charging / Discharging", ""],
            ["Hex2Str", 1, "Alarm - todo", ""],
            ["BigHex2Short:r/1000", 2, "Voltage Difference", "V"],
            ["BigHex2Short:r/1000", 2, "Balance Current", "A"],
            ["BigHex2Short:r/1000", 2, "Balance Trigger Voltage", "V"],
            ["BigHex2Short:r/1000", 2, "Max Balance Current", "A"],
            ["Hex2Str", 1, "Balance On / Off", ""],
            ["Hex2Int", 1, "Set Number of Cells", ""],
            ["BigHex2Short:r/1000", 2, "Voltage Cell01", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell02", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell03", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell04", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell05", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell06", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell07", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell08", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell09", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell10", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell11", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell12", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell13", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell14", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell15", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell16", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell17", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell18", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell19", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell20", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell21", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell22", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell23", "V"],
            ["BigHex2Short:r/1000", 2, "Voltage Cell24", "V"],
            ["BigHex2Short", 2, "Temperature", "°C"],
            ["Hex2Str", 1, "Checksum", ""],
        ],
        "test_responses": [
            bytes.fromhex(
                "EB 90 01 FF 1E D3 0F 69 14 13 02 00 00 00 07 00 00 00 05 03 E8 01 14 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 0F 69 00 16 6F"
            ),
        ],
    },
}


class jk485(jkAbstractProtocol):
    def __str__(self):
        return "JKBMS RS485 serial communication protocol handler"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        self._protocol_id = b"JK485"
        self.COMMANDS = COMMANDS
        self.STATUS_COMMANDS = [
            "getBalancerData",
        ]
        self.SETTINGS_COMMANDS = [
            "",
        ]
        self.DEFAULT_COMMAND = "getBalancerData"

    def get_full_command(self, command) -> bytes:
        """
        Override the default get_full_command as its different for JK485
        """
        log.info(f"Using protocol {self._protocol_id} with {len(self.COMMANDS)} commands")
        # These need to be set to allow other functions to work`
        self._command = command
        self._command_defn = self.get_command_defn(command)
        # End of required variables setting
        if self._command_defn is None:
            # Maybe return a default here?
            return None
        if "command_code" in self._command_defn:
            # full command is 7 bytes long
            cmd = bytearray(7)
            # 55 AA 01 FF 00 00 FF
            # frame header  0x55 0xaa
            cmd[0:2] = bytes.fromhex("55aa")
            log.debug(f"cmd with header: {cmd}")
            # slave address 0x01
            cmd[2] = 0x01
            log.debug(f"cmd with header + slave address: {cmd}")
            # command code  0xff
            cmd[3] = int(self._command_defn["command_code"], 16)
            # frame data    0x0000
            cmd[4:6] = bytes.fromhex("0000")
            log.debug(f"cmd with command code and frame data: {cmd}")
            # checksum      0xff
            cmd[-1] = crc8(cmd)
            log.debug(f"cmd with crc: {cmd}")
            return cmd



import logging
from enum import auto

from strenum import LowercaseStrEnum

from powermon.formats import getFormatfromConfig, DEFAULT_FORMAT

# Set-up logger
log = logging.getLogger("outputs")

DEFAULT_OUTPUT = "screen"


class OutputType(LowercaseStrEnum):
    SCREEN = auto()
    MQTT = auto()
    API_MQTT = auto()


def getOutputClass(output_type, formatter, output_config={}):
    output_class = None
    # Only import the required class
    log.debug("outputType %s" % output_type)
    if output_type == OutputType.MQTT:
        from powermon.outputs.mqtt import MQTT
        #add from config here
        output_class = MQTT.from_config(output_config)
        output_class.set_formatter(formatter)
        #output_class = MQTT(output_config=output_config, mqtt_broker=mqtt_broker, formatter=formatter)
    elif output_type == OutputType.API_MQTT:
        from powermon.outputs.api_mqtt import ApiMqtt
        output_class = ApiMqtt.from_config(output_config)
        output_class.set_formatter(formatter)
    else:
        from powermon.outputs.screen import Screen
        output_class = Screen.from_config(output_config)
        output_class.set_formatter(formatter)
    return output_class


def getOutputs(outputsConfig):
    # outputs can be None,
    # str (eg screen),
    # list (eg [{'type': 'screen', 'format': 'simple'}, {'type': 'screen', 'format': {'type': 'htmltable'}}])
    # dict (eg {'format': 'table'})
    # print("outputs %s, type: %s" % (outputs, type(outputs)))
    _outputs = []
    log.debug("processing outputsConfig: %s", outputsConfig)
    if outputsConfig is None:
        _outputs.append(parseOutputConfig({"type": "screen", "format": "simple"}))
    elif isinstance(outputsConfig, str):
        # eg 'screen'
        _outputs.append(parseOutputConfig(outputsConfig))
    elif isinstance(outputsConfig, list):
        # eg [{'type': 'screen', 'format': 'simple'}, {'type': 'screen', 'format': {'type': 'htmltable'}}]
        for outputConfig in outputsConfig:
            _outputs.append(parseOutputConfig(outputConfig))
    elif isinstance(outputsConfig, dict):
        # eg {'format': 'table'}
        _outputs.append(parseOutputConfig(outputsConfig))
    else:
        pass
    return _outputs


#TODO: clean up the multiple types of outputconfig, there should only be one type
def parseOutputConfig(outputConfig):
    log.debug("parseOutputConfig, config: %s", outputConfig)
    topic_override = None
    # outputConfig can be None, a str (eg 'screen') a dict (eg {'format': 'table'}) or a list (eg [{'type': 'screen', 'format': 'simple'}])
    if outputConfig is None:
        log.debug("got blank outputConfig")
        outputType = DEFAULT_OUTPUT
        _format = getFormatfromConfig(DEFAULT_FORMAT)
        log.debug("got format: %s", (_format))
        _output = getOutputClass(outputType, formatter=_format)
        log.debug("got output: %s", (_output))
        return _output
    elif isinstance(outputConfig, str):
        # eg 'screen'
        log.debug("got str type outputConfig: %s", outputConfig)
        outputType = outputConfig
        _format = getFormatfromConfig(DEFAULT_FORMAT)
        log.debug("got format: %s", (_format))
        _output = getOutputClass(outputType, formatter=_format)
        log.debug("got output: %s", (_output))
        return _output
    elif isinstance(outputConfig, dict):
        # eg {'format': 'table'}
        log.debug("got dict type outputConfig: %s", outputConfig)
        outputType = outputConfig.get("type", DEFAULT_OUTPUT)
        formatConfig = outputConfig.get("format", DEFAULT_FORMAT)

        _format = getFormatfromConfig(formatConfig)
        log.debug("got format: %s", _format)
        _output = getOutputClass(outputType, formatter=_format)
        log.debug("got output: %s", _output)
        return _output
    elif isinstance(outputConfig, list):
        # eg [{'type': 'screen', 'format': 'simple'}], possibly multiple outputs
        # loop through outputs
        log.debug("got list type outputConfig: %s", outputConfig)
        raise Exception("Problem in outputs init - list type for outputConfig")

        # for item in outputConfig:
        #     _out = parseOutputConfig(item, topic, schedule_name, device, mqtt_broker)
        #     if _out is not None:
        #         outputs.append(_out)
        # return outputs
    else:
        # miss configured output?
        log.warning("outputConfig (%s) doesnt match expectations, defaulting", outputConfig)
        outputType = DEFAULT_OUTPUT
        formatConfig = DEFAULT_FORMAT

    _format = getFormatfromConfig(formatConfig)
    log.debug("got format: %s", (_format))



from pydantic import BaseModel, Field
#from powermon.dto.response_definition_dto import ResponseDefinitionDTO


class CommandDefinitionDTO(BaseModel):

    command_code: str
    description: str
    help_text: str | None
    result_type: str | None
    #responses: list | None
    #test_responses : list[bytes] | None
    regex : str | None
